#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from serv.test_httppool import getPayload, check_response
from fdi.dataset.testproducts import get_demo_product, get_related_product
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.product import Product
from fdi.dataset.dateparameter import DateParameter
from fdi.dataset.serializable import serialize
from fdi.pal.context import MapContext
from fdi.pal.urn import parseUrn
from fdi.dataset.arraydataset import ArrayDataset
from fdi.dataset.stringparameter import StringParameter
from fdi.pal.productstorage import ProductStorage
from fdi.dataset.eq import deepcmp
from fdi.pns.jsonio import commonheaders, auth_headers
from fdi.dataset.classes import Classes, Class_Module_Map
from svom.products.projectclasses import Class_Look_Up
from fdi.utils.getconfig import getConfig
from fdi.utils.common import lls
from conftest import csdb_pool_id, SHORT
from fdi.pal.publicclientpool import PublicClientPool
from fdi.pns.public_fdi_requests import read_from_cloud, load_from_cloud
from fdi.utils.getconfig import getConfig
from fdi.pns.fdi_requests import reqst

import json
import time
import os
from functools import lru_cache
from datetime import datetime
import pytest
import requests
import logging
# create logger

logging.basicConfig(level=logging.DEBUG,
                    format='%(asctime)s %(name)8s %(process)d '
                    '%(threadName)s %(levelname)s %(funcName)10s()'
                    '%(lineno)3d- %(message)s')

logger = logging.getLogger(__name__)
logger.debug('logging level %d' % (logger.getEffectiveLevel()))

pc = getConfig()

# markers


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "cmp_wipe"
    )

# -------------------TEST CSDB WITH Requests.session a client -----
# low level and not as complete as the later part.


@pytest.fixture(scope='session')
def csdb_client(urlcsdb):
    urlupload = urlcsdb + '/datatype/upload'
    urldelete = urlcsdb + '/datatype/'
    urllist = urlcsdb + '/datatype/list'
    client = requests.session()
    headers = client.headers
    headers = auth_headers(pc['cloud_username'], pc['cloud_password'],
                           headers)
    client.headers.update(headers)
    yield urlupload, urldelete, urllist, client


def get_all_prod_types(urllist, client):
    x = client.get(urllist)
    o, code = getPayload(x)
    types = o['data']
    return types


def add_a_dataType(full_name, jsn, client, urlup):
    """ not using csdb_client fixture. returns csdb ['data'] result. """
    hdr = {"accept": "*/*"}
    fdata = {"file": (full_name, jsn)}
    data = {"metaPath": "/metadata", "productType": full_name}
    x = reqst(client.post, apis=urlup, files=fdata,
              data=data, headers=hdr)
    return x


asci = False


@lru_cache(maxsize=8)
def cls2jsn(clsn):
    obj = Class_Look_Up[clsn]()
    return json.dumps(obj.zInfo, ensure_ascii=asci, indent=2)


def pool_exists(poolname, csdb_c, urlc, create_clean=False):
    urlupload, urldelete, urllist, client = csdb_c

    x = client.get(urlc+f"/pool/info?storagePoolName={poolname}")
    assert x.status_code == 200
    o, code = getPayload(x)
    if o['code'] == 0:
        if create_clean:
            for u in get_all_in_pool(poolname, what='urn', urlc=urlc, client=client, nulltype=False):
                part = u[3:].replace(':', '/')
                x = client.post(urlc+f'/storage/delete?path={part}')
                assert x.status_code == 200, x.text
                o, code = getPayload(x)
                if o['code']:
                    raise RuntimeError(
                        'Cleanin exisiting pool {part} failed.')
        return True
    elif create_clean:
        x = client.post(
            urlc+f"/pool/create?poolName={poolname}&read=0&write=0")
        assert x.status_code == 200
        o, code = getPayload(x)
        return o['code'] == 0
    return o['code'] == 0


def upload_defintion(full_cls, urllist, urlupload, urldelete,
                     client=None, check=True):
    """upload the definition of given class.

    """
    if isinstance(full_cls, list):
        alist = True
        fs = full_cls
    else:
        alist = False
        fs = [full_cls]

    if check:
        # new one not there?
        types = get_all_prod_types(urllist, client=client)
        for full_cls in fs:
            if full_cls in types:  # and full_name.startswith('sv.'):
                logger.debug('confirm dataTypes'+lls(types, 80))

    # upload
    for full_cls in fs:
        jsn = cls2jsn(full_cls.rsplit('.', 1)[-1])
        r = add_a_dataType(full_cls, jsn, client=client, urlup=urlupload)
        assert r.status_code == 200
    if check:
        # check ptypes again
        types = get_all_prod_types(urllist, client)
        # print(types)
        for full_cls in fs:
            assert full_cls in types
    return fs if alist else fs[0]


USE_SV_MODULE_NAME = False
Tx = 'TP' + '_0X'  # + str(datetime.now())


def test_upload_def_Tx(csdb_client):
    """ define a prodect and upload the definition """
    urlupload, urldelete, urllist, client = csdb_client

    cls = Tx
    if USE_SV_MODULE_NAME:
        cls_full_name = f'sv.{cls}'  # cls_full_name.rsplit('.', 1)[-1]
    else:
        cls_full_name = Class_Module_Map[cls] + '.' + cls
    upload_defintion(cls_full_name, urllist, urlupload, urldelete,
                     client=client)


def upload_prod_data(prd, cls_full_name,
                     desc, obj, obs_id, instr, start, end,
                     level='CL1a', program='GRB', url='',
                     client=None, verify=False, token='',
                     asyn=False, **kwds):
    """ upload product data """

    hdr = client.headers
    hdr.update(kwds.pop('header', {}))

    prd.description = desc
    prd.instrument = StringParameter(instr)
    prd.version = '0.1'
    prd.startDate = DateParameter(start)
    prd.startDate = DateParameter(end)
    prd.meta['object'] = StringParameter(obj)
    prd.meta['obs_id'] = StringParameter(obs_id)
    prd.level = level
    prd.meta['program'] = StringParameter(program)
    jsn = prd.serialized(indent=2)

    if 0:
        filen = f'/cygdrive/d/code/tmp/clz/{cls_full_name}.jsn'
        with open(filen, 'w+', encoding='utf-8') as f:
            f.write(jsn)
        hdr = {"accept": "*/*", 'X-CSDB-METADATA': '/_ATTR_meta'}
        with open(filen, 'rb') as f:
            fdata = [("file", (filen, f))]
    if 1:
        fdata = {'file': (cls_full_name, jsn)}
        data = {'tags': 'foo,'+str(datetime.now())}

        hdr.update({"accept": "*/*", 'X-CSDB-METADATA': '/_ATTR_meta'})
        hdr['X-AUTH-TOKEN'] = token
        hdr['X-CSDB-AUTOINDEX'] = '1'
        hdr['X-CSDB-METADATA'] = '/_ATTR_meta'
        hdr['X-CSDB-HASHCOMPARE'] = '0'

        # url urlupdata = urlcsdb + f'/storage/{pool}/{cls_full_name}'
        # x = client.post(url, files=fdata, data=data, headers=hdr)
        if asyn:
            if 1:

                res = load_from_cloud('uploadProduct', token=token,
                                      products=jsn,
                                      path=['/%s/%s' %
                                            tuple(url.rsplit('/', 2)[-2:])],
                                      tags=str(datetime.now()),
                                      resourcetype=cls_full_name,
                                      header=hdr,
                                      asyn=True,
                                      **kwds)
                od = res[0]
                assert len(od) > 4
            else:
                x = reqst('post', [url], data=data, headers=hdr,
                          **kwds)
                assert x.status_code == 200, lls(x.text, 500)
        else:
            x = reqst(client.post, url, files=fdata,
                      data=data, headers=hdr, **kwds)
            assert x.status_code == 200, lls(x.text, 500)
            o, code = getPayload(x)
            od = o['data']
            assert o['code'] == 0
        urn = od['urn']
        path = od['path']
        typ = od['dataType']

        if not typ:
            logger.warning(
                f'{urn} has no dataType. Upload Datatype definition to fix.')
        logger.debug(f"uploaded product urn={urn} path={path} type={typ}")
        """ output example:
    Example:
    {
      "code": 0,
      "data": {
        "dataType": "string",
        "index": 0,
        "md5": "string",
        "path": "string",
        "size": 0,
        "tags": [
          "string"
        ],
        "timestamp": 0,
        "type": "string",
        "url": "string",
        "urn": "string"
      },
      "msg": "string",
      "total": 0
    }
"""
    if verify:
        # read back
        y = client.get(od['url'])
        assert y.text == jsn
        logger.debug('upload verified')
    logger.info('UPLD URN %s' % od['urn'])
    return od


def ddd():
    if 1:
        x = add_a_dataType(cls_full_name, jsn=jsn, client=client)
        #### !!!! This throws error !!! ####
        assert x.status_code == 200, x.text
        o, code = getPayload(x)
        assert o['code'] == 0
        assert o['data'] is None


def test_upload_data_Tx(csdb_client, urlcsdb, csdb_token):
    """ upload a Tx prod. definition not auto uploaded"""

    urlupload, urldelete, urllist, client = csdb_client

    pool = csdb_pool_id
    cls = Tx
    if USE_SV_MODULE_NAME:
        cls_full_name = f'sv.{cls}'
    else:
        cls_full_name = Class_Module_Map[cls] + '.' + cls

    logger.debug(f'Upload {cls_full_name} data. Datatype %s found in type list.' % (
        '' if cls_full_name in get_all_prod_types(urllist, client=client) else 'not'))
    urlupdata = urlcsdb + f'/storage/{pool}/{cls_full_name}'
    prod = Class_Look_Up[cls]()
    res = upload_prod_data(prod,
                           cls_full_name=cls_full_name,
                           desc='Demo_Product',
                           obj='3c273',
                           obs_id='b2000a',
                           instr='VT',
                           start='2000-01-01T00:00:00',
                           end='2001-01-01T00:00:00',
                           level='CL2a',
                           program='PPT',
                           url=urlupdata,
                           client=client,
                           verify=True,
                           token=csdb_token,
                           asyn=ASYN
                           )
    logger.debug(f'{cls_full_name} uploaded {res}')


def get_all_in_pool(poolname=None, path=None, what='urn', urlc='', client=None, limit=10000, nulltype=True):
    """Return all of something in a pool.

    Parameters
    ----------
    poolname : str
         name of te pool. If both `poolname` and `path` are given, take `poolname`.
    path : str
        part in a path '{poolname}', or '/{poolname}/{product-name}', or
        '{poolname}/{product-name}/{index.aka.serial-number}',
        e.g. '/sv1/sv.BaseProduct'
    what : str
        which item in ['data'] list to return. e.g. 'urn' means a list
        of URNs found in the poolname. 'count' means the length of
        ['data'].
    urlc : str
        path to be prepended before `/storage/searcBy...`.
    client : Session
        requets.Session
    limit : str
        how many items per page maximum.

    Returns
    -------
    list or int
        list of `what` or lenth.

    """
    url = urlc
    if poolname:
        url += f'/storage/searchByPoolOrPath?limitCount={limit}&pool={poolname}'
    else:
        url += f'/storage/searchByPoolOrPath?limitCount={limit}&path={path}'

    x = client.get(url)
    assert x.status_code == 200
    o, code = getPayload(x)
    assert o['code'] == 0
    od = o['data']
    prodList = od if nulltype else [p for p in od if p['dataType']]
    """
  "data": [
    {
      "url": "http://123.56.102.90:31702/csdb/v1/storage/sv1/fdi.dataset.testproducts.TP/22",
      "path": "/sv1/fdi.dataset.testproducts.TP/22",
      "urn": "urn:sv1:fdi.dataset.testproducts.TP:22",
      "timestamp": 1673396754678,
      "tags": [],
      "index": 22,
      "md5": "A69D8B3411999CF25500969B53B691BA",
      "size": 45395,
      "contentType": null,
      "fileName": "fdi.dataset.testproducts.TP",
      "dataType": "fdi.dataset.testproducts.TP"
    }, ...]
    """
    return len(prodList) if what == 'count' else [p[what] for p in prodList]


def test_getDataInfo(urlcsdb, csdb_uploaded, csdb_client):  # _uploaded
    # test_pool, url, pstore = csdb
    test_pool, uniq, resPrds, pstore = csdb_uploaded
#    r = resPrds
    pname = test_pool.poolname

    # add another prod in pool2
    pool2 = test_pool._poolname + '2'
    cls = 'TCC'
    if USE_SV_MODULE_NAME:
        cls_full_name = f'sv.{cls}'
    else:
        cls_full_name = Class_Module_Map[cls] + '.' + cls
    aip_ctl = add_a_prod_in_another_pool(
        pool2, urlcsdb, cls_full_name, csdb_client, nulltype=True)

    assert len(aip_ctl) > 0

    # this only has non-null types
    pinfo = test_pool.getPoolInfo()[pname]

    tget = test_pool.getDataInfo

    # info of all prods in this pool
    dinfo = tget()
    assert dinfo == tget('') == tget(what='', paths='', pool=pname)
    assert len(tget(nulltype=False)) == test_pool.getCount()

    # get all URNs in the pool. including nulltype
    urns = tget('urn')
    assert set(urns) == set(pinfo['_urns'])
    assert set(x['urn'] for x in dinfo) == set(pinfo['_urns'])
    # of all pools
    p2urns = tget('urn', pool=pool2)
    assert set(p2urns) == set(aip_ctl)

    # not include null typed. and all pools
    urns2 = tget('urn', paths=cls, pool='', nulltype=False)
    p2urns2 = tget('urn', pool=pool2, nulltype=False)
    p3urns2 = tget('urn', pool=pname+'3', nulltype=False)
    # this may fail if other pools have
    assert set(urns2) >= set(
        u for u in pinfo['_urns'] if cls in u) | set(p2urns2) | set(p3urns2)

    # get info of a URN with a identifying part as `paths`
    if urns2:
        assert tget('', urns2[0][3:].replace(':', '/'))[0]['urn'] == urns2[0]

    # a list of urns: get their index one by one
    urn_test_pool = [x for x in urns2 if pname in x]
    num = min(len(urn_test_pool), 5)
    t0 = time.time()
    ilsns = []
    _urns = urn_test_pool[:num]
    for part in _urns:
        if pname not in part:
            continue
        sn = part.rsplit(':', 1)[-1]
        isn = int(sn)
        assert tget('index', part, nulltype=False)[0] == isn
        ilsns.append([isn])

    logger.info(time.time() - t0)

    # a list of urns: get their indices as a list
    t0 = time.time()
    parts = list(map(lambda x: x[3:].replace(':', '/'), urn_test_pool[:num]))
    # take only the first if multiple are found
    inds = [i[:1] for i in tget('index', parts)]
    # stupid csdb behavior: if sn is 0, sn is not shown.
    # p = [[0] if i == [] else i for i in ilsns]
    p = ilsns
    assert inds == p
    logger.info(f'no asyn get index {time.time() - t0}')

    # a list of urns: fix the pool, get their indices as a list
    t0 = time.time()
    po, dt, sn = parseUrn(urns[:num])
    parts = list(map(lambda x: x[3:].replace(':', '/'), urn_test_pool[:num]))
    # take only the first if multiple are found
    inds = [i[:1] for i in tget('index', parts, asyn=1)]
    logger.info(f'asyn {time.time() - t0}')


def delDataTypeData(urlc, poolname, cls_full_name, client):
    """ This API deletes all products of the given datatype and the datatype object. """
    logger.debug('Try storage/delDatatype')
    url = urlc + \
        f"/storage/delDatatype?path=/{poolname}/{cls_full_name}"
    x = client.delete(url)
    logger.debug('x.text: '+x.text)
    if x.status_code != 200:
        return False
    o, code = getPayload(x)
    assert o['code'] == 0, o['msg']
    return o


def delete_datatype(urlc, cls_full_name, client):
    """ This API deletes the datatype object of the given datatype. """
    logger.debug('Try delete /v1/datatype/dataType')
    url = urlc + \
        f"/datatype/{cls_full_name}"
    x = client.delete(url)

    logger.debug('x.text: '+text)
    if x.status_code != 200:
        return x.text
    o, code = getPayload(x)
    assert o['code'] == 0, o['msg']
    return o


def delete_defintion(clsn, full_name, urllist, urldelete,
                     client=None, poolname='', urlc='', deldatt=True):
    """delete the definition of a given class.

    Datatypes are not supposed to be deleted frequently. Use upload to update.
    """

    types = get_all_prod_types(urllist, client=client)
    # logger.info(lls(types, 80), full_name in types)

    # The first step is different depending on whether full_name is in
    if full_name in types:
        logger.debug(
            f'Type {full_name} exists in type list. Use upload to update. Datatypes are not supposed to be deleted frequently.')
        if deldatt:
            res = delDataTypeData(urlc, poolname, full_name, client=client)
            if not res:
                # delDataTypeData failed, usually due to no exisiting products
                res = delete_datatype(urlc, full_name, client=client)
        # gone
        types = get_all_prod_types(urllist, client)
        assert full_name not in types, ''
    else:
        logger.debug(f'Type {full_name} not found in type list. Skip.')


# @pytest.mark.skip
def test_del_def_Tx(csdb_client, csdb):
    """ define a product and upload/update the definition. WARNING: not for regular use. """
    urlupload, urldelete, urllist, client = csdb_client
    test_pool, url, pstore = csdb
    urlc = urldelete.rsplit('/datatype', 1)[0]

    cls = Tx
    if USE_SV_MODULE_NAME:
        cls_full_name = f'sv.{cls}'  # cls_full_name.rsplit('.', 1)[-1]
    else:
        cls_full_name = Class_Module_Map[cls] + '.' + cls

    pinfo = test_pool.getPoolInfo()
    pname = test_pool.poolname

    # add type
    jsn = cls2jsn(cls)
    x = add_a_dataType(cls_full_name, jsn=jsn,
                       client=client, urlup=urlupload)
    clz = test_pool.getProductClasses()
    assert cls_full_name in clz
    delete_defintion(cls, cls_full_name, urllist, urldelete,
                     client=client, poolname=pname, urlc=urlc, deldatt=True)


def xxxfollow():
    # add a type and a prod
    x = add_a_dataType(cls_full_name, jsn=jsn,
                       client=client, urlup=urlupload)
    assert cls_full_name in clz
    clz = test_pool.getProductClasses()
    assert cls_full_name in clz
    urlupdata = urlc + f'/storage/{pname}/{cls_full_name}'
    res = upload_prod_data(prod,
                           cls_full_name=cls_full_name,
                           desc='Demo_Product',
                           obj='3c273',
                           obs_id='b2000a',
                           instr='VT',
                           start='2000-01-01T00:00:00',
                           end='2001-01-01T00:00:00',
                           level='CL2a',
                           program='PPT',
                           url=urlupdata,
                           client=client,
                           verify=True
                           )
    assert res['type'] == cls_full_name
    prod_urn = res['urn']
    # Then delete
    with pytest.raises(TypeError):
        delete_defintion(cls, cls_full_name, urllist, urldelete,
                         client=client, urlc=urlc, deldatt=True)
    pinfo = test_pool.getPoolInfo()
    # The prod is still there
    assert prod_urn in pinfo[pname]['_urns']


@ pytest.fixture(scope='function')
def upload_7products(csdb, urlcsdb, csdb_client, tmp_prods):
    """ Upload a number of products.

    Returns
    -------
    list
       Example `[{'index': 647,
  'md5': '4F15BDC99BFDB1077F5EF80A98924094',
  'path': '/test_csdb_fdi/fdi.dataset.testproducts.TB/647',
  'size': 9838,
  'tags': ['foo', '2023-01-28 11:15:33.938743'],
  'timestamp': 1674875733148,
  'dataType': None,
  'url': 'http://...:31702/csdb/v1/storage/test_csdb_fdi/fdi.dataset.testproducts.TB/647',
  'urn': 'urn:test_csdb_fdi:fdi.dataset.testproducts.TB:647'}, ...]`

    """
    test_pool, poolurl, pstore = csdb
    urlupload, urldelete, urllist, client = csdb_client
    # get_all_prod_types(urlcsdb+'/datatype/list', ftest_pool.client)
    # 7 products
    prds = tmp_prods

    all_data = []
    for i, prd in enumerate(prds):
        cls = prd.__class__.__name__
        # make full names
        if USE_SV_MODULE_NAME:
            cls_full_name = f'sv.{cls}'  # cls_full_name.rsplit('.', 1)[-1]
        else:
            cls_full_name = Class_Module_Map[cls] + '.' + cls
        urlupdata = urlcsdb + f'/storage/{test_pool.poolname}/{cls_full_name}'
        all_data.append(upload_prod_data(prd,
                                         cls_full_name=cls_full_name,
                                         desc=f'Demo_{cls}',
                                         obj=cls,
                                         obs_id=f'b2000a_{cls_full_name}',
                                         instr='VT',
                                         start='2000-01-01T00:00:00',
                                         end='2001-01-01T00:00:00',
                                         level='CL2a',
                                         program='PPT',
                                         url=urlupdata,
                                         client=client
                                         )
                        )
    return all_data


def test_upload_products(csdb_7types_defined, upload_7products):
    all_types = csdb_7types_defined
    all_data = upload_7products

    assert len(all_data) == 7
    assert set(d['dataType'] for d in all_data) == {t for t in all_types}


def add_a_prod_in_another_pool(poolname2, urlc, cls_full_name, csdb_c, nulltype=True):
    path2 = f'{poolname2}/{cls_full_name}'
    urlupdata = urlc + f'/storage/{path2}'
    cls = cls_full_name.rsplit('.', 1)[1]
    prd = Classes.mapping[cls]()
    pool_exists(poolname2, csdb_c, urlc,
                create_clean=True)
    jsn = cls2jsn(cls)
    add_a_dataType(cls_full_name, jsn, csdb_c[-1], urlc+'/datatype/upload')
    upload_prod_data(prd,
                     cls_full_name=cls_full_name,
                     desc=f'delte_dataType_control',
                     obj='ctrl',
                     obs_id=f'test_ctrl_{cls_full_name}',
                     instr='VT',
                     start='2000-01-01T00:00:00',
                     end='2001-01-01T00:00:00',
                     level='CL2a',
                     program='PPT',
                     url=urlupdata,
                     client=csdb_c[-1]
                     )
    # number of control prod
    aip_ctl = get_all_in_pool(
        path=f'/{path2}', what='urn', urlc=urlc, client=csdb_c[-1],
        nulltype=nulltype
    )
    assert all(cls_full_name in a for a in aip_ctl)
    return aip_ctl


@ pytest.fixture(scope='function')
def get_list(csdb_client, urlcsdb):
    """ list all urns in a pool """

    urlupload, urldelete, urllist, client = csdb_client
    urlc = urlcsdb
    allurns = get_all_in_pool(csdb_pool_id, None, 'urn', urlc, client)
    # logger.info(pformat(allurns))
    return allurns


def test_del_7products(urlcsdb, csdb_client, upload_7products):
    """ delete product data from a pool"""

    urlupload, urldelete, urllist, client = csdb_client
    poolname = csdb_pool_id
    data = upload_7products
    urlc = urlcsdb

    l0 = get_all_in_pool(
        path=f'/{poolname}', what='urn', urlc=urlc, client=client)

    # 1 for del one peod; 0 for del all of the type
    del1 = 1
    urns = []
    for dt in data:
        # ‘dataType' in post storage/pool/prod, None in ['type']
        cls_full_name = dt['path'].rsplit('/', 2)[-2]
        ind = dt['index']
        path = f'/{poolname}/{cls_full_name}'
        # the first urn in aip does not have  ':0' at the end
        urn = dt['urn'][:-2] if dt['urn'].endswith(':0') else dt['urn']
        if del1:  # delete one product
            # number of prod of the same type
            aip = get_all_in_pool(
                path=path, what='urn', urlc=urlc, client=client)
            assert urn in aip
            # delete
            url = urlc + \
                f"/storage/delete?path=/{path}/{ind}"
            x = client.post(url)
            aip = get_all_in_pool(
                path=path, what='urn', urlc=urlcsdb, client=client)
            assert urn not in aip

        else:
            url = urlc + \
                f"/storage/delete?path=/{path}/{ind}"
            x = client.post(url)

            if 0:
                # add a prod in a different pool {pool}2
                aip_ctl = add_a_prod_in_another_pool(
                    poolname+'2', urlc, cls_full_name, csdb_client)
                # delete from pool 1
                urlxxx = urlc + \
                    f"/storage/delDatatypeData?path=/{poolname}/{cls_full_name}"
                url = urlc + \
                    f"/storage/delete?path=/{poolname}/{cls_full_name}{ind}"
                x = client.delete(url)

        assert x.status_code == 200, x.text
        o, code = getPayload(x)
        assert o['code'] == 0, o['msg']
        urns.append(urn)
        if del1:
            del1 = 0
    # fixture get_list is exhausted so use get_all_in_pool
    l1 = get_all_in_pool(
        path=f'/{poolname}', what='urn', urlc=urlc, client=client)
    assert set(l0) - set(l1) == set(urns)
    assert set(l0) == set(l1) | set(urns)
    if 0:
        aip_ctl2 = get_all_in_pool(
            path=f'/{poolname}2/{cls_full_name}', what='urn', urlc=urlc, client=client)
        # control pool not affected
        assert len(aip_ctl) == len(aip_ctl2)


# ----------------------TEST CSDB WITH ProductStorage----------------


def genProduct(size=1, cls='ArrayDataset', unique='', prod_cls=None):
    res = []
    if prod_cls is None:
        prod_cls = Product
    for i in range(size):
        x = prod_cls(description="product example with several datasets" + unique,
                     instrument="Crystal-Ball", modelName="Mk II", creator='Cloud FDI developer')
        i0 = i
        i1 = [[i0, 2, 3], [4, 5, 6], [7, 8, 9]]
        i2 = 'ev'  # unit
        i3 = 'image1'  # description
        image = ArrayDataset(data=i1, unit=i2, description=i3)
        # put the dataset into the product
        x["RawImage"] = image
        x.set('QualityImage', ArrayDataset(
            [[0.1, 0.5, 0.7], [4e3, 6e7, 8], [-2, 0, 3.1]]))
        res.append(x)
    if size == 1:
        return res[0]
    else:
        return res


def genMapContext(size=1):
    map1 = MapContext(description='product with refs 1')
    map1['creator'] = 'Cloud FDI developer'
    return map1


@ pytest.fixture(scope="session")
def csdb_token(csdb, pc):
    logger.info('test token')
    test_pool, url, pstore = csdb

    token = pc['cloud_token']
    if token != test_pool.token:
        logger.info("Tokens are not equal or not synchronized")
    return test_pool.token


def test_csdb_token(csdb_token):
    tok = csdb_token


def test_csdb_createPool(csdb_new):
    logger.info('test create a brand new pool')
    test_pool, url, pstore = csdb_new
    try:
        pinfo = test_pool.poolInfo[test_pool.poolname]
        assert len(pinfo['_classes']) == 0
    except ValueError:
        # found deleted pool by this name.
        assert test_pool.restorePool() is True
    assert test_pool.poolExists() is True
    assert test_pool.poolExists('non_existant') is False


def test_csdb_poolInfo(csdb):
    test_pool, url, pstore = csdb
    test_pool.getPoolInfo()
    # logger.info(test_pool.poolInfo)


def test_clean_csdb(clean_csdb):
    logger.info('test get classes')
    test_pool, url, pstore = clean_csdb
    pinfo = test_pool.getPoolInfo()
    assert len(pinfo[test_pool.poolname]['_classes'].keys()) == 0
    assert len(pinfo[test_pool.poolname]['_tags'].keys()) == 0
    assert len(pinfo[test_pool.poolname]['_urns'].keys()) == 0


def test_getProductClasses(urlcsdb, csdb, csdb_client):
    urlupload, urldelete, urllist, client = csdb_client
    test_pool, url, pstore = csdb
    clz = test_pool.getProductClasses()
    # add another prod in pool2
    pool2 = test_pool._poolname + '000'
    cls = 'TP_0X'
    if USE_SV_MODULE_NAME:
        cls_full_name = f'sv.{cls}'
    else:
        cls_full_name = Class_Module_Map[cls] + '.' + cls
    aip_ctl = add_a_prod_in_another_pool(
        pool2, urlcsdb, cls_full_name, csdb_client)

    assert len(aip_ctl) > 0
    assert all(cls_full_name not in c for c in clz)
    assert cls_full_name in aip_ctl[-1]


@ pytest.fixture(scope=SHORT)
def csdb_7types_defined(csdb, urlcsdb, csdb_client, tmp_prod_types):
    """
    Returns
    -------
    list
        all full dataType names.
    """

    urlupload, urldelete, urllist, client = csdb_client
    ftest_pool, poolurl, pstore = csdb
    # get_all_prod_types(urlcsdb+'/datatype/list', ftest_pool.client)
    # 7 products
    prd_types, seri = tmp_prod_types

    asci = False

    all_data = []
    for i, ty in enumerate(prd_types):
        cls = ty.__name__
        # make full names
        if USE_SV_MODULE_NAME:
            full_cls = f'sv.{cls}'  # cls_full_name.rsplit('.', 1)[-1]
        else:
            full_cls = Class_Module_Map[cls] + '.' + cls
        all_data.append(full_cls)
    defns = upload_defintion(all_data,
                             urllist, urlupload, urldelete,
                             client=client, check=1)
    logger.debug(f'Added/updated definition for {defns}.')

    assert len(prd_types) == len(defns)
    return defns


@ pytest.fixture(scope='function')
def csdb_uploaded(csdb, csdb_7types_defined, csdb_client):
    return csdb_up(csdb, csdb_7types_defined, csdb_client, asyn=ASYN, ntimes=1)


# total of 7*N
N = 3
ASYN = 0


@ pytest.fixture(scope='function')
def csdb_uploaded_n(csdb, csdb_7types_defined, csdb_client):
    return csdb_up(csdb, csdb_7types_defined, csdb_client, asyn=ASYN, ntimes=N)


def csdb_up(_csdb, _csdb_7types_defined, _csdb_client, ntimes, asyn=False):
    urlupload, urldelete, urllist, client = _csdb_client
    test_pool, poolurl, pstore = _csdb
    poolname = test_pool._poolname

    all7Types = _csdb_7types_defined

    cnt0 = test_pool.getCount(update=True)
    logger.info(
        f'To upload all7Types ...count={cnt0}')
    t0 = time.time()

    uniq = str(t0)
    resPrds = []

    for i, full_cls in enumerate(all7Types):
        cls = full_cls.rsplit('.', 1)[1]
        ptype = Classes.mapping[cls]
        prds = [ptype(f'demo {cls} {uniq}')] * ntimes
        r = pstore.save(prds, asyn=asyn)
        # assert all(i for i in r), full_cls
        resPrds.extend(r)

    n0 = len(resPrds)
    assert n0 == len(all7Types) * ntimes
    # pinfo = test_pool.getPoolInfo()
    logger.info(
        f'Generated in {time.time()-t0} secs n0={n0}, count={test_pool.getCount(update=True)}')
    # logger.info(f'{test_pool.getDataInfo("_urns")}')
    return test_pool, uniq, resPrds, pstore


def test_csdb_upload(csdb_uploaded):
    logger.info('test upload multiple products')
    test_pool, uniq, resPrds, pstore = csdb_uploaded
    r = resPrds

    urninfo = test_pool.getDataInfo('urn')
    assert len(r) == 7
    assert len(urninfo) >= 7

    for ele in resPrds:
        assert csdb_pool_id in ele.urn
        assert ele.urn in urninfo
        assert uniq in ele.product.description

    urnsN7 = [ele.urn for ele in resPrds] * N
    po, pt, sn = parseUrn(urnsN7)
    po = test_pool._poolname
    ures = []
    oldCount = test_pool.getCount(update=True)
    assert len(urninfo) == oldCount

    tags = []
    prds = []
    for n, ref in enumerate([*resPrds] * N):
        tags.append(f'{n} {pt[n]}')
        prds.append(ref.product)
    t0 = time.time()
    if not ASYN:
        for n, ref in enumerate([*resPrds] * N):
            ures.append(pstore.save(prds[n], tag=tags[n],
                        poolname=po))
    else:
        rs = pstore.save(prds, tag=tags, poolname=po, asyn=ASYN)
        ures = rs

    cnt1 = test_pool.getCount()
    assert cnt1 == oldCount + N*7
    logger.info(f'Uploaded in {time.time()-t0} {len(ures)}, curr cnt {cnt1}')


def test_csdb_load_Prd(csdb_uploaded):
    logger.info('test load product')
    # test_pool, url, pstore = csdb

    test_pool, uniq, resPrds, pstore = csdb_uploaded
    pinfo = test_pool.getPoolInfo()
    urns = list(r.urn for r in resPrds)
    us = list(pinfo[test_pool.poolname]['_urns'].keys())
    for u in urns:
        prd = pstore.load(u).product
        assert prd.description.endswith(uniq), 'retrieve production incorrect'


@ pytest.fixture(scope='function')
def csdb_addTag(csdb_uploaded):
    logger.info('test add tag to urn')
    # test_pool, url, pstore = csdb

    test_pool, uniq, resPrds, pstore = csdb_uploaded
    pinfo = test_pool.getPoolInfo()
    ppu = pinfo[test_pool._poolname]['_urns']
    times = str(int(time.time()))
    tag = 'test_prd_' + times

    # get a tag-free URN
    for u, ts in reversed(ppu.items()):
        if len(ts) == 0:
            break
    urn = u

    assert not test_pool.tagExists(tag)
    test_pool.setTag(tag, urn)
    assert test_pool.tagExists(tag)
    with pytest.raises(AssertionError):
        assert tag in test_pool.getTags(urn)
    assert tag in test_pool.getTags(urn, update=True)
    # pinfo = test_pool.getPoolInfo()

    # multiple tags
    tag1 = 'test_prd1_' + times
    tag2 = ['test_prd2', 'test_prd3']
    # get a tag-free URn
    start = 0
    for u, ts in reversed(ppu.items()):
        if not start:
            if u == urn:
                start = 1
            continue
        if len(ts) == 0:
            break
    urn2 = u
    assert not test_pool.tagExists(tag1)

    test_pool.setTag(tag1, urn2)
    test_pool.setTag(tag2, urn2)
    tagsall = [tag1]+tag2
    tu2 = test_pool.getTags(urn2)
    with pytest.raises(AssertionError):
        assert set(tu2) == set(tagsall)
    # add update=1
    tu2 = test_pool.getTags(urn2, update=True)
    assert set(tu2) == set(tagsall)
    return test_pool, tag, urn, tagsall, urn2


def test_csdb_delTag(csdb_addTag):
    logger.info('test delete a tag')

    test_pool, tag, urn, tag2, urn2 = csdb_addTag
    assert tag in test_pool.getTags(urn)
    tag_urn_map = test_pool.getTagUrnMap()
    assert tag_urn_map is not None

    test_pool.removeTag(tag)
    test_pool.getPoolInfo()
    assert tag not in test_pool.getTags(urn)
    assert tag2[0] in test_pool.getTags(urn2)
    assert tag2[1] in test_pool.getTags(urn2)

    test_pool.removeTag(tag2[1])
    test_pool.getPoolInfo()
    assert tag2[1] not in test_pool.getTags(urn2)
    assert tag2[0] in test_pool.getTags(urn2)


def test_csdb_count(csdb_uploaded):
    logger.info('test count')
    # test_pool, url, pstore = csdb

    # start with none-empty
    test_pool, uniq, resPrds, pstore = csdb_uploaded
    poolname = test_pool.poolname

    pinfo = test_pool.getPoolInfo()
    for clazz, cld in pinfo[poolname]['_classes'].items():
        assert test_pool.getCount(clazz) == \
            len(pinfo[poolname]['_classes'][clazz]['sn'])

    if 0:
        raise ValueError("pool i empty. can't get count()")

    # add prods again.
    prds = Class_Look_Up[clazz.rsplit('.', 1)[-1]]('getCount extra')
    resPrds2 = pstore.save([prds])
    # pinfo = test_pool.getPoolInfo()
    assert len(resPrds2) == 1

    last_cnt = len(pinfo[poolname]['_classes'][clazz]['sn'])
    count = test_pool.getCount(clazz)
    assert count - last_cnt == 1


def test_csdb_remove(csdb_uploaded):
    logger.info('test remove product')
    # test_pool, url, pstore = csdb
    test_pool, uniq, resPrds, pstore = csdb_uploaded
    poolname = test_pool.poolname
    pinfo = test_pool.getPoolInfo()
    cls = resPrds[0].getType().__name__

    # test remove with URN
    if USE_SV_MODULE_NAME:
        cls_full_name = f'sv.{cls}'
    else:
        cls_full_name = Class_Module_Map[cls] + '.' + cls
    rdIndex = pinfo[poolname]['_classes'][cls_full_name]['sn'][-1]
    urn = 'urn:' + csdb_pool_id + ':' + cls_full_name + ':' + str(rdIndex)
    res = test_pool.remove(urn)
    assert res == 0, res
    pinfo = test_pool.getPoolInfo()
    assert cls_full_name not in pinfo[poolname]['_classes'] or rdIndex not in pinfo[poolname]['_classes'][cls_full_name]['sn']
    # test pstore
    urn = list(pinfo[poolname]['_urns'])[-1]
    res = pstore.remove(urn)
    assert res == 0, res
    pinfo = test_pool.getPoolInfo()
    assert urn not in pinfo[poolname]['_urns']


NWIPE = N


def test_csdb_wipe_(csdb_uploaded_n, csdb_client):
    logger.info('test wipe all')
    # test_pool, url, pstore = csdb

    test_pool, uniq, resPrds, pstore = csdb_uploaded_n
    urlupload, urldelete, urllist, client = csdb_client
    pname = test_pool.poolname

    if 1:
        oldCount = test_pool.getCount(update=True)
        n0 = oldCount  # , NWIPE*7)
        logger.info(f'wipe setup {n0}')
        t0 = time.time()
        path = f"/{pname}"
        # res = read_from_cloud('remove', token=test_pool.token,
        #                      path=path)
        rec = test_pool.removeAll()

    elif 0:
        tyinfo = test_pool.getProductClasses(count=True)
        assert tyinfo
        assert isinstance(tyinfo, set), str(tyinfo)
        pinfo = test_pool.getPoolInfo(update_hk=True)
        # logger.info(list(pinfo[pname]['_urns'])[-8:])

        oldCount = sum(test_pool.getCount(typename=t, update=False)
                       for t in tyinfo)
        n0 = oldCount  # , NWIPE*7)
        logger.info(f'wipe setup, {n0}')
        t0 = time.time()

        # first not delDataTypeData
        for cls in tyinfo:
            path = f"/{pname}/{cls}"
            res = read_from_cloud('delDataTypeData', token=test_pool.token,
                                  path=path)
        # test_pool.removeAll(ignore_error=True)
    newCount = test_pool.getCount(update=True)
    assert newCount == 0
    logger.info(f'>>> {time.time()-t0} {oldCount} -> {newCount}')


what_wipe = pytest.mark.skipif(0, reason='digging')


@what_wipe
def test_cmp_wipe1(csdb_client, csdb_uploaded_n):
    test_pool, uniq, resPrds, pstore = csdb_uploaded_n
    urlupload, urldelete, urllist, client = csdb_client
    pname = test_pool.poolname
    info = test_pool.getPoolInfo(update_hk=True)
    assert isinstance(info, dict), str(info)
    oldCount = test_pool.getCount()
    assert oldCount
    n0 = len(info[pname]['_urns'])
    logger.info(f'wipe setup {n0}, {oldCount}')
    t0 = time.time()

    for clazz, cld in info[pname]['_classes'].items():
        # all datatypes
        # types = get_all_prod_types(urllist, client)
        # assert clazz in types
        path = f'/{pname}/{clazz}'
        res = read_from_cloud('delDataTypeData', token=test_pool.token,
                              path=path)
    logger.info(
        f"delDataTypeData, >>> {time.time()-t0}, {n0}, {len(info[pname]['_urns'])}")
    assert test_pool.getCount(update=True) == 0


DTYPE = 'fdi.dataset.testproducts.TP'

ONLY_IN_test_pool = True
""" products of this type are going to be removed in test_pool, not all csdb storage."""

if ONLY_IN_test_pool:
    @ what_wipe
    def test_cmp_wipe2(csdb, csdb_client, csdb_uploaded_n):
        wipe_all(_csdb_client=csdb_client,
                 _csdb_uploaded_n=csdb_uploaded_n)
else:
    @ what_wipe
    def test_wipe_urns(csdb, csdb_client):
        wipe_all(_csdb_client=csdb_client, _csdb=csdb)


def wipe_all(_csdb=None, _csdb_client=None, _csdb_uploaded_n=None):
    # products of this type are going to be removed in csdb storage.

    if ONLY_IN_test_pool:
        test_pool, uniq, resPrds, pstore = _csdb_uploaded_n
    else:
        test_pool, poolurl, pstore = _csdb
    urlupload, urldelete, urllist, client = _csdb_client
    urlc = urldelete.rsplit('/datatype', 1)[0]
    if ONLY_IN_test_pool:
        pname = test_pool.poolname
        test_pool._poolname = pname
        info = test_pool.getPoolInfo()

        oldCount = test_pool.getCount()
        assert oldCount
        assert isinstance(info, dict), str(info)
        urns = info[pname]['_urns']
        assert len(urns) == oldCount
    else:
        path_part = DTYPE
        _all = get_all_in_pool(
            poolname=None, path=path_part, what='urn', urlc=urlc, client=client
        )
        urns = [u for u in _all if DTYPE in u]
        oldCount = len(urns)
    n0 = min(len(urns), NWIPE*7)
    logger.info(f'wipe setup, {n0}, {lls(urns, 300)}')
    # pp = input(f'wipe2 {n0} URNs?')
    if len(urns) > 999:
        urns = urns[:999]
    t0 = time.time()
    res = test_pool.remove(list(urns), asyn=ASYN)
    logger.info(f'urn >>> {time.time()-t0}, n0={n0}, len={len(urns)}')
    newCount = max(oldCount - n0, 0)
    assert test_pool.getCount(update=True) == newCount

    ##########


def test_cmp_wipe3(csdb_client, csdb_uploaded_n):
    test_pool, uniq, resPrds, pstore = csdb_uploaded_n
    urlupload, urldelete, urllist, client = csdb_client
    urlc = urldelete.rsplit('/datatype', 1)[0]
    pname = test_pool.poolname
    test_pool._poolname = pname
    info = test_pool.getPoolInfo()
    oldCount = test_pool.getCount()
    assert oldCount

    n0 = min(oldCount, NWIPE*7)
    logger.info(f'wipe setup, {n0}.')
    t0 = time.time()
    test_pool.ignore_error_when_delete = True
    for clazz, cld in info[pname]['_classes'].items():
        res = test_pool.remove(resourcetype=clazz, index=cld['sn'])
    logger.info(
        f"'class-sn', >>> {time.time()-t0}, {n0}, {len(info[pname]['_urns'])}")
    newCount = max(oldCount - n0, 0)
    assert test_pool.getCount(update=True) == newCount



# -*- coding: utf-8 -*-

from .deserialize import deserialize
from ..utils.getconfig import get_mqtt_config
from .listener import EventListener, EventSender
from .serializable import serialize

import paho.mqtt.client as mqtt
from paho.mqtt.client import error_string, MQTT_ERR_NO_CONN

from itertools import chain
import logging

if 0:
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(name)8s %(process)d %(threadName)s %(levelname)s %(funcName)10s() %(lineno)3d- %(message)s')

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class MqttRelayListener(EventListener):
    """ Generic interface for sending anything heard to an MQTT message queue.
    """

    def __init__(self, topics=None,
                 host=None, port=None, username=None, passwd=None,
                 callback=None, clean_session=None,
                 client_id=None, userdata=None,
                 qos=1,
                 conn=True, subs=True,
                 **kwds):  # MqttRelayListener
        """ Starts a MQTT message queue and forward everything in the arguement list to the MQ serialized.

        host, port, username, passwd: if any is not provided, it is looked up in `config[...].
        """
        super().__init__(**kwds)

        if bool(host and port and username and passwd) is False:
            conf = get_mqtt_config()

        mq = mqtt.Client(
            client_id=client_id,
            clean_session=clean_session,
            userdata=userdata if userdata else self)

        mq.username_pw_set(username if username else conf['mq_user'],
                           passwd if passwd else conf['mq_pass'])

        self.mq = mq
        self.topics = topics

        # for topic subscription
        if isinstance(self.topics, list):
            if isinstance(self.topics[0], str):
                # topics is a list of topics
                topics = [(topic, qos) for topic in self.topics]
            else:
                topics = self.topics
        elif isinstance(self.topics, str):
            topics = self.topics
        else:
            logger.error('Bad format for subscrib() topics ' +
                         str(self.topics))
            return None
        self.topics_for_subscription = topics

        self.host = host if host else conf['mq_host']
        self.port = port if port else int(conf['mq_port'])
        self.qos = qos

        self.keepalive = True
        self.username = username if username else conf['mq_user']
        self.passwd = passwd if passwd else conf['mq_pass']
        mq.username_pw_set(self.username, self.passwd)

        #mq.on_message = callback if callback else on_message
        mq.on_connect = on_connect

        self.mq.loop_start()
        # connect
        if conn:
            mq.connect(self.host, self.port, self.keepalive)
            logger.debug("Connect " + self.host + ":" + str(self.port))

    def targetChanged(self,  *args, **kwargs):
        """ Informs that an event has happened in a target of
        any type.
        """

        payload = list(chain(args, kwargs.items()))
        json_str = serialize(payload)

        logger.debug("send msg to [" + self.topics + "]")
        logger.debug(json_str)

        self.mq.reconnect()
        msg_info = self.mq.publish(
            self.topics, payload=json_str, qos=self.qos, retain=False)
        rc, mid = msg_info.rc, msg_info.mid

        if rc == MQTT_ERR_NO_CONN:
            raise Exception('why not connected?')
        logger.debug("Publish status: %d mid %d" % (rc, mid))
        logger.debug("send over")


class MqttRelaySender(EventSender):
    """ Gets MQTT messages and forwards to listeners.

    """

    def __init__(self, topics=None,
                 host=None, port=None, username=None, passwd=None,
                 callback=None, clean_session=None,
                 client_id=None, userdata=None, keepalive=60,
                 qos=1, **kwds):
        """ Starts a MQTT message queue and forward everything in the arguement list to the MQ serialized.

        host, port, username, passwd: if any is not provided, it is looked up in `config[...].
        """
        super().__init__(**kwds)

        if bool(host and port and username and passwd) is False:
            conf = get_mqtt_config()
        logger.debug('starting mq listening to '+str(topics))
        mq = mqtt.Client(
            client_id=client_id,
            clean_session=clean_session,
            userdata=userdata if userdata else self)

        username = username if username else conf['mq_user']
        passwd = passwd if passwd else conf['mq_pass']
        mq.username_pw_set(username, passwd)
        host = host if host else conf['mq_host']
        port = port if port else int(conf['mq_port'])
        mq.on_message = on_message
        mq.connect(host, port, keepalive=keepalive)
        logger.debug("Connect " + host + ":" + str(port))
        rc, mid = mq.subscribe(topics, qos=qos)
        logger.debug("subscribe %s status: %s mid %s" %
                     (str(topics), error_string(rc), str(mid)))
        self.mq = mq

        mq.loop_start()


def on_connect(client, userdata, flags, rc):
    has_prev_data = ' some' if flags['session present'] else ' no'
    msg = mqtt.connack_string(rc)  # CONN_RESULT[rc]
    logmsg = "Connected with result: " + msg + \
        has_prev_data + ' prev session data.'

    rc, mid = userdata.mq.subscribe(userdata.topics_for_subscription,
                                    qos=userdata.qos)
    logger.debug("subscribe %s status: %s mid %s" %
                 (str(userdata.topics), error_string(rc), str(mid)))


def on_message(client, userdata, msg):
    mqtt_rel_s = userdata
    logger.debug("Received: " + msg.topic + ' ' + str(msg.payload))

    msgobj = deserialize(msg.payload.decode(encoding='utf-8'))

    mqtt_rel_s.fire(msgobj)
    mqtt_rel_s.last_msg = msgobj



from fdi.utils.fits_kw import FITS_KEYWORDS,getFitsKw
from fdi.dataset.arraydataset import ArrayDataset
from astropy.io import fits

import numpy as np
import os

def main():
    fitsdir = '/Users/jia/desktop/vtse_out/'
    if os.path.exists(fitsdir + 'array.fits'):
        os.remove(fitsdir + 'array.fits')
    ima=ArrayDataset(data=[[1,2,3,4],[5,6,7,8]], description='a')
    imb=ArrayDataset(data=[[1,2,3,4],[5,6,7,8],[1,2,3,4],[5,6,7,8]], description='b')
    #im=[[1,2,3,4],[5,6,7,8]]
    hdul = fits.HDUList()
    fits_dataset(hdul,[ima,imb])

def fits_dataset(hdul,ima_list):
    
    
    for n, ima in enumerid(ima_list):
        
        
        header = fits.Header()
        if issubclass(ima.__class__, ArrayDataset):
            a=np.array(ima)
            print (a.shape)
            hdul.append(fits.imageHDU(a),header=header)
            header=add_header(a.meta,header)

        elif issubclass(ima.__class__, TableDataset):
            units=[]
            dtype=[]
            data=[getRow(Slice)]
            desc=[]
            for name, col in ima.items():
                units.append(col.unit)
                dtype.append(col.typecode)
                
                desc.append(col.description)
            t=table(name=ima.getColumnNames(), units=units, dtype=dtype, data=data)
            
            hdul.append(fits.BinTableHDU(a),header=header)
            header=add_header(a.meta,header)
            
        elif issubclass(ima.__class__, CompositeDataset):
            dataset=fits_dataset(hdul,ima)
    return hdul

    #hdul.writeto(fitsdir + 'array.fits')
    
   # f=fits.open(fitsdir + 'array.fits')
   # print(len(f))
    #h1=f[0].header
    #h2=f[1].header
    #print(h2)
    #return h1

def add_header(meta,header):
    for name,param in meta.items():
        kw=getFitsKw(name)
        header[kw]=(param.value,param.description)
    return header
   
def fits_header():
    fitsdir = '/Users/jia/desktop/vtse_out/'
    f=fits.open(fitsdir + 'array.fits')
    h=f[0].header
    #h.set('add','header','add a header')
    h['add']=('header','add a header')
    h['test']=('123','des')
    f.close()
    
    return h    
    
def test_fits_kw(h):
    #print(h)
    print(list(h.keys()))
    #assert FITS_KEYWORDS['CUNIT'] == 'cunit'
    assert getFitsKw(list(h.keys())[0]) == 'SIMPLE'
    assert getFitsKw(list(h.keys())[3]) == 'NAXIS1'

if __name__ == '__main__':

    #test_fits_kw(fits_data())
    main()


# -*- coding: utf-8 -*-

from . import productref
from .poolmanager import PoolManager
from .productpool import ProductPool, makeLockpath
from .urn import Urn
from ..dataset.odict import ODict

import filelock
from weakref import finalize

import logging
# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))


class ProductStorage(object):
    """ Logical store created from a pool or a poolURL.

    Every instanciation with the same pool will  result in a new instance of ProdStorage.

    """

    def __init__(self, pool=None, poolurl=None, poolmanager=None,
                 **kwds):
        """ Gets the storage "control pannel" for pool with specifed name.

        `auth` with `client`, can be given here to be passed to `PoolManager.getPool`.
        `client` how to call remote api if poolurl indicates a remote pool. Default is `None` for using the configured host/port/credentials. If doing a mocked server test, this needs to be set.
        :pool: can be given as the only pramemter. If `auth` and `client` are given they will substitute those of  `pool`. If `pool` is not given, those will need to be given.
        :poolname: if is a string will be taken as a poolname. if is a pool object will be registered with its name,
        :poolurl: is sent to the PoolManager with poolname to get the pool object.

        """
        if issubclass(pool.__class__, str) and ':' in pool:
            raise TypeError(
                'First argument must be a poolname or a pool object, not ' + str(pool))
        super(ProductStorage, self).__init__()

        if poolmanager:
            # poolservers have their own PoolManagers
            self.PM = poolmanager
        else:
            from .poolmanager import PoolManager
            self.PM = PoolManager

        self._pools = ODict()  # dict of poolname - poolobj pairs
        self.register(pool=pool, poolurl=poolurl, **kwds)

    def register(self,  poolname=None, poolurl=None, pool=None,
                 **kwds):
        """ Registers the given pools to the storage.

        :client: passed to `PoolManager.getPool`.
        :auth: passed to `PoolManager.getPool`.
        """

        if issubclass(pool.__class__, str) and poolname is None:
            pool, poolname = poolname, pool
        with filelock.FileLock(makeLockpath('ProdStorage', 'w')), \
                filelock.FileLock(makeLockpath('ProdStorage', 'r')):
            if pool and issubclass(pool.__class__, ProductPool):
                _p = self.PM.getPool(pool=pool, **kwds)
            elif poolurl is None and poolname is None:
                # quietly return for no-arg construction case
                return
            else:
                if poolname is not None and not issubclass(poolname.__class__, str):
                    raise TypeError('Poolname must be a string, not ' +
                                    poolname.__class__.__name__)
                if poolurl is not None and not issubclass(poolurl.__class__, str):
                    raise TypeError('Poolurl must be a string, not ' +
                                    poolurl.__class__.__name__)
                _p = self.PM.getPool(poolname=poolname, poolurl=poolurl,
                                     **kwds)
            self._pools[_p._poolname] = _p

        logger.debug('registered pool %s -> %s.' %
                     (str(pool), str(self._pools)))

    def unregister(self, pool=None, ignore_error=False, **kwds):
        """ Unregisters the given pools to the storage.
        """

        with filelock.FileLock(makeLockpath('ProdStorage', 'w')):
            if issubclass(pool.__class__, ProductPool):
                poolname = pool.getId()
            else:
                poolname = pool
            if self.PM.isLoaded(poolname):
                # remove frpm pool manager
                # TODO i dentify self
                res = self.PM.remove(poolname, ignore_error=ignore_error)
                # do this after del above
                del self._pools[poolname]
                logger.debug('unregistered pool %s -> %s.' %
                             (str(pool), str(self._pools)))
            else:
                logger.info('Pool %s is not registered.' % poolname)
        return

    def unregisterAll(self, ignore_error=False):
        self.PM.removeAll(ignore_error=ignore_error)
        self._pools.clear()

    def load(self, urnortag):
        """ Loads a product with a URN or a list of products with a tag, from the (writeable) pool.

        It always creates new ProductRefs.
        :return: productref if there is only one. A ```list``` of ```ProductRefs```.
        urnortag: urn or tag
        """
        poolname = self.getWritablePool()

        def runner(urnortag):
            if issubclass(urnortag.__class__, list):
                ulist = list(map(runner, urnortag))
                return ulist
            else:
                if issubclass(urnortag.__class__, str):
                    if len(urnortag) > 3 and urnortag[0:4].lower() == 'urn:':
                        urns = urnortag
                    else:
                        urns = self.getUrnFromTag(urnortag)
                        ret = []
                        for x in urns:
                            pr = productref.ProductRef(
                                urn=x, poolname=poolname)
                            ret.append(pr)
                        return ret
                elif issubclass(urnortag.__class__, Urn):
                    urns = urnortag.urn
                else:
                    raise ValueError(
                        'must provide urn, urnobj, tags, or lists of them')
                return productref.ProductRef(urn=urns, poolname=poolname)
        ls = runner(urnortag=urnortag)
        # return a list only when more than one refs
        return ls  # if len(ls) > 1 else ls[0]

    def save(self, product, tag=None, poolname=None, geturnobjs=False, asyn=False, **kwds):
        """ saves to the writable pool if it has been registered.

        Parameters
        ----------
        product : BaseProduct, list
            Product or a list of them or '[ size1, prd, size2, prd2, ...]'.
        tag : str, list
            If given a tag, all products will be having this tag.
        If a list tags are given to every one product then the
        number of tags must not be the same to that of `product`. If
        they are equal, each tag is goven to the product at the same
        index in the `product` list.
        serialize_out : bool
            if `True` returns contents in serialized form.
        serialize_in : bool
            If set, product input is serialized.
        poolName: str
            If the named pool is not registered, registers and saves.
        geturnobjs : bool
            returns UrnObjs if geturnobjs is True.
        kwds: options passed to json.dump() for subclasses.

        Returns
        -------
        ProductRef: Product reference.
        Urn: If `geturnobjs` is set.
        str: If `serialze_out` is set, serialized form of `ProductRef` or `URN`.
        list: `list` of the above of input is a list.
        """

        if poolname is None:
            if len(self._pools) > 0:
                poolname = self.getWritablePool()
            else:
                raise ValueError('no pool registered')
        elif poolname not in self._pools:
            self.register(poolname)

        if logger.getEffectiveLevel() <= logging.DEBUG:
            desc = [x.description[-6:] for x in product] if issubclass(
                product.__class__, list) else product.description[-6:]
            logger.debug('saving product:' + str(desc) +
                         ' to pool ' + str(poolname) + ' with tag ' + str(tag))

        try:
            ret = self._pools[poolname].saveProduct(
                product, tag=tag, geturnobjs=geturnobjs,
                asyn=asyn, **kwds)
        except Exception as e:
            logger.error('unable to save to the writable pool.')
            raise
        from fdi.pal.productref import ProductRef
        if issubclass(ret.__class__, list):
            if ret and issubclass(ret[0].__class__, ProductRef):
                for r in ret:
                    r.setStorage(self)
        return ret

    def remove(self, urn, datatype=None, index=None, ignore_error=False):
        """ removes product of urn from the writeable pool
        """
        poolname = self.getWritablePool()
        logger.debug('removing product:' + str(urn) +
                     ' from pool ' + str(poolname))
        try:
            pool = self._pools[poolname]
            pool.ignore_error_when_delete = ignore_error
            res = pool.remove(urn, resourcetype=datatype, index=index)
            return res
        except Exception as e:
            msg = 'unable to remove from the writable pool.'
            if ignore_error:
                logger.error(msg)
                return None
            else:
                raise

    def accept(self, visitor):
        """ Hook for adding functionality to object
        through visitor pattern."""
        visitor.visit(self)

    def getHead(self, ref):
        """ Returns the latest version of a given product, belonging
        to the first pool where the same track id is found.
        """
        raise NotImplementedError()

    def getPools(self):
        """  Returns the set of ProductPools registered.
        mh: in a list of poolnames
        XXX TODO: getPoolnames
        """
        return list(self._pools.keys())

    def getPool(self, poolname):
        """ mh: returns the pool object by poolname from this storage
        """
        if poolname not in self._pools:
            msg = 'pool ' + poolname + ' not found'
            logger.error(msg)
            raise NameError(msg)
        return self._pools[poolname]

    def getWritablePool(self, obj=False):
        """ returns the poolname of the first pool, which is the only writeable pool.

        :obj: (bool) return the pool objject instead of the name.
        """
        l = list(self._pools.items())

        return l[0][1] if obj else l[0][0]

    def getAllTags(self):
        """ Get all tags defined in the writable pool.
        """
        return self._pools[self.getWritablePool()].getTags()

    def getProductClasses(self, poolname):
        """  Yields all Product classes found in this pool.
        """
        return self._pools[poolname].getProductClasses()

    def getTags(self, urn):
        """  Get the tags belonging to the writable pool that associated to a given URN.
        returns an iterator.
        """
        return self._pools[self.getWritablePool()].getTags(urn)

    def getMeta(self, urn):
        """  Get the metadata belonging to the writable pool that associated to a given URN.
        returns an ODict.
        """
        if not issubclass(urn.__class__, str):
            urn = urn.urn

        return self._pools[self.getWritablePool()].meta(urn)

    def getUrnFromTag(self, tag):
        """ Get the URN belonging to the writable pool that is associated
        to a given tag.
        """

        return self._pools[self.getWritablePool()].getUrn(tag)

    def wipePool(self, ignore_error=False, asyn=False, **kwds):
        """ Clear all data and meta data of the writable pool.
        """

        self.ignore_error_when_delete = ignore_error
        pool = self.getWritablePool(obj=True)
        pool.ignore_error_when_delete = ignore_error
        pool.removeAll(
            ignore_error=ignore_error, asyn=asyn, **kwds)

    def isEmpty(self):
        """ Returns whether all pools are empty or there is no pool. """

        return len(self._pools) == 0 or all(p.isEmpty() for p in self._pools.values())

    def select(self, query, variable=None, ptype=None, previous=None):
        """ Returns a list of URNs to products that match the specified query.

        Parameters
        ----------
        query : the query object, or str
            The Query instances or
             the 'where' query string to make a query object.
        variable : str
            name of the dummy variable in the query string.
            if `variable` is 'm', query goes via `MetaQuery(ptype, query)` ; else by `AbstractQuery(ptype, variable, query)` .
        ptype : class
            The class object whose instances are to be queried. Or
            fragment of the name of such classes.
        previous : list or str
            of urns, possibly from previous search. or a string of comma-separated urns, e.g. `'urn:a:foo:12,urn:b:bar:9'`

        Returns
        -------
        the set of return eferences to products matching the supplied query.
        """
        ret = []
        if issubclass(previous.__class__, str):
            previous = previous.split(',')
        if issubclass(query.__class__, str) and ptype and variable:
            # search all registered pools
            for poolnm, pool in self._pools.items():
                ret += pool.select(query, variable, ptype, previous)
            return ret

        # search all registered pools
        for poolnm, pool in self._pools.items():
            ret += pool.select(query, previous)
        return ret

    def __getstate__(self):
        """ Can be encoded with serializableEncoder """
        return OrderedDict(writablePool=self.getWritablePool())

    def __repr__(self):
        return self.__class__.__name__ + '( pool= ' + str(self._pools if hasattr(self, '_pools') else None) + ' )'



# -*- coding: utf-8 -*-

from .getswag import swag
from .. import checkpath, PM_S
from .httppool_server import (
    excp,
    resp
)
# from .. import auth
from ..model.user import auth

from ..._version import __version__
from ...dataset.deserialize import deserialize_args, deserialize
from ...pal.productpool import PoolNotFoundError
from ...pal.webapi import WebAPI
# from ...pal.urn import parseUrn
from ...pal.dicthk import HKDBS
from ...utils.getconfig import getConfig
from ...utils.common import (lls,
                             logging_ERROR,
                             logging_WARNING,
                             logging_INFO,
                             logging_DEBUG
                             )
from flask import Blueprint, jsonify, request, current_app, url_for, abort
from werkzeug.exceptions import HTTPException
# from flasgger import swag_from

import shutil
import time
import copy
import os
from os.path import join, expandvars
from itertools import chain


endp = swag['paths']

pools_api = Blueprint('pools', __name__)


######################################
####   get_pools_url   ####
######################################


@ pools_api.route('', methods=['GET'])
def get_pools_url():
    """ Get names and urls of all pools, registered or not.
    """
    logger = current_app.logger

    ts = time.time()
    path = current_app.config['FULL_BASE_LOCAL_POOLPATH']
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Listing all directories from ' + path)

    result = get_name_all_pools(path)

    burl = request.base_url
    if issubclass(result.__class__, list):
        res = dict((x, '/'.join((burl, x))) for x in result)
    else:
        res = {}

    dvers = getConfig('docker_version')
    svers = getConfig('server_version')

    from ..model.user import getUsers
    # report lru_cache info
    lru = 'LRU user '
    for name, usr in current_app.config['USERS'].items():
        lru += '%s: %s ' % (name, usr.getCacheInfo())

    msg = '%d pools found. Versiosn: fdi %s docker %s pool server %s' % (
        len(result), __version__, dvers, svers)
    msg += lru
    code = 200
    return resp(code, res, msg, ts)


######################################
#### /  get_pools   ####
######################################


@ pools_api.route('/', methods=['GET'])
def get_pools():
    """ Get names of all pools, registered or not.
    """
    logger = current_app.logger

    ts = time.time()
    path = current_app.config['FULL_BASE_LOCAL_POOLPATH']
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Listing all directories from ' + path)

    result = get_name_all_pools(path)

    res = result
    msg = '%d pools found.' % len(result)
    code = 200
    return resp(code, res, msg, ts)


def get_name_all_pools(path):
    """ Returns names of all pools in the given directory.

    """

    logger = current_app.logger
    alldirs = []
    os.makedirs(path, exist_ok=True)
    allfilelist = os.listdir(path)
    for file in allfilelist:
        filepath = join(path, file)
        if os.path.isdir(filepath):
            alldirs.append(file)
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(path + ' has ' + lls(alldirs, 100))
    return alldirs


####################################################
#### /pools   get_all_pools_size_counts/        ####
#### 333#######################################V#####


@ pools_api.route('/pools', methods=['GET'])
@ auth.login_required(role=['read_only', 'read_write'])
def get_all_pools_size_counts():
    """ Returns a list of Pool IDs (pool names) of all pools and their size and file counts.
    ---
    """

    logger = current_app.logger
    ts = time.time()
    serialize_out = False
    FAILED = '"FAILED"' if serialize_out else 'FAILED'
    data_path = current_app.config['FULL_BASE_LOCAL_POOLPATH']
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Information of all  pools.')

    if not os.path.exists(data_path):
        return resp(404, FAILED, f'Find no directory at {data_path}')
    pools = {}
    s = c = 0
    for i in os.listdir(data_path):
        _, count, msg, size = get_data_count(None, i)
        pools[i] = {'size_in_bytes': size, 'number_of_data_files': count}
        s += size
        c += count
    result = pools
    msg = 'There is/are %d pools holding %d products. %.1d MB.' % \
        (len(pools), c, s/1000000)
    code = 200
    return resp(code, result, msg, ts)

##########################################
#### /pools/   get_registered_pools/  ####
##########################################


@ pools_api.route('/pools', methods=['GET'])
@ pools_api.route('/pools/', methods=['GET'])
@ auth.login_required(role=['read_only', 'read_write'])
def get_registered_pools():
    """ Returns a list of Pool IDs (pool names) of all pools registered with the Global PoolManager.
    ---
    """

    logger = current_app.logger
    ts = time.time()
    path = current_app.config['FULL_BASE_LOCAL_POOLPATH']
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Listing all registered pools.')

    # [p.getPoolurl() for p in PM_S.getMap()()]
    result = list(PM_S.getMap())
    msg = 'There is/are %d pools registered to the PoolManager.' % len(result)
    code = 200
    return resp(code, result, msg, ts)


######################################
#### /pools/register_all pools/register_all/  ####
######################################


@ pools_api.route('/pools/register_all', methods=['PUT'])
@ pools_api.route('/pools/register_all/', methods=['PUT'])
@auth.login_required(role='read_write')
def register_all():
    """ Register (Load) all pools on tme server.


    """

    logger = current_app.logger
    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Load all pools.')
    # will have / at the end
    burl = request.base_url.rsplit('register_all', 1)[0]
    result, bad = load_pools(None, auth.current_user())
    code = 400 if len(bad) else 200
    # result = ', '.join(pmap.keys())
    if issubclass(result.__class__, dict):
        result = dict((x, burl+x) for x in result)
    msg = '%d pools successfully loaded. Troubled: %s' % (
        len(result), str(bad))
    return resp(code, result, msg, ts)


def load_pools(poolnames, usr):
    """
    Adding all pool to server pool storage.

    poolnames: if given as a list of poolnames, only the exisiting ones of the list will be loaded.
    :usr: current authorized user.
    Returns: a `dict` of successfully loaded pools names-pool in `good`, and troubled ones in `bad` with associated exception info.
    """

    logger = current_app.logger
    path = current_app.config['FULL_BASE_LOCAL_POOLPATH']
    pmap = {}
    bad = {}
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('loading all from ' + path)
    alldirs = poolnames if poolnames else get_name_all_pools(path)
    for nm in alldirs:
        # must save the link or PM_S._GlobalPoolList will remove as dead weakref
        code, thepool, msg = register_pool(nm, usr=usr)
        if code == 200:
            pmap[nm] = thepool
        else:
            bad[nm] = nm+': '+msg

    if logger.isEnabledFor(logging_DEBUG):
        logger.debug("Registered pools: %s in %s" %
                     (str(list(pmap.keys())), path))
    return pmap, bad

######################################
#### /pools/unregister_all  pools/unregister_all/  ####
######################################


@ pools_api.route('/pools/unregister_all', methods=['PUT'])
@ pools_api.route('/pools/unregister_all/', methods=['PUT'])
@ auth.login_required(role='read_write')
def unregister_all():

    logger = current_app.logger
    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('unregister-all ' + pool)

    good, bad = unregister_pools()
    code = 200 if not bad else 416
    result = good
    msg = '%d pools unregistered%s' % (len(good),
                                       (' except %s.' % str(bad) if len(bad) else '.'))
    return resp(code, result, msg, ts)


def unregister_pools(poolnames=None):
    """
    Removing all pools from the PoolManager.
`w
    poolnames: if given as a list of poolnames, only the ones in the list will be unregistered.

    Returns: a list of successfully unregistered pools names in `good`, and troubled ones in `bad` with associated exception info.
    """
    logger = current_app.logger

    good = []
    notgood = []
    all_pools = poolnames if poolnames else copy.copy(
        list(PM_S.getMap().keys()))
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('unregister pools ' + str(all_pools))

    for nm in all_pools:
        code, res, msg = unregister_pool(nm)
        if res == 'FAILED':
            notgood.append(nm+': '+msg)
        else:
            good.append(nm)
    return good, notgood

######################################
#### /pools/wipe_all  pools/wipe_all/  ####
######################################


@ pools_api.route('/pools/wipe_all', methods=['DELETE'])
# @ pools_api.route('/pools/wipe_all/', methods=['DELETE'])
@ auth.login_required(role='read_write')
def wipe_all():
    """ Remove contents of all pools.

    Only registerable pools will be wiped. Pool directories are not removed.
    """
    logger = current_app.logger
    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Wipe-all pool')

    good, bad = wipe_pools(None, auth.current_user())
    code = 200 if not bad else 416
    result = good
    msg = '%d pools wiped%s' % (len(good),
                                (' except %s.' % str(bad) if len(bad) else '.'))
    return resp(code, result, msg, ts)


def wipe_pools(poolnames, usr):
    """
    Deleting all pools using pool api so locking is properly used.

    poolnames: if given as a list of poolnames, only the  ones in the list will be deleted.

    Returns: a list of successfully removed pools names in `good`, and troubled ones in `bad` with associated exception info.
    """
    logger = current_app.logger
    path = current_app.config['FULL_BASE_LOCAL_POOLPATH']
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('DELETING pools contents from ' + path)

    # alldirs = poolnames if poolnames else get_name_all_pools(path)

    good = []
    notgood = []
    all_pools, not_loadable = load_pools(poolnames, usr)
    names = list(all_pools.keys())
    for nm in copy.copy(names):
        thepool = all_pools[nm]
        try:
            thepool.removeAll()
            shutil.rmtree(join(path, nm))
            res = PM_S.remove(nm)
            if res > 1:
                notgood.append(nm+': '+str(res))
                logger.info('Pool %s not deleted.' % nm)
            else:
                good.append(nm)
                logger.info('Pool %s deleted.' % nm)
        except Exception as e:
            notgood.append(nm+': '+str(e))
    return good, notgood

######################################
####    /{pool}    /{pool}/  GET  ####
######################################


@ pools_api.route('/<string:pool>/', methods=['GET'])
@ pools_api.route('/<string:pool>', methods=['GET'])
# @ auth.login_required(role=['read_only', 'read_write'])
def get_pool(pool):
    """ Get information of the given pool.

    Returns the state of the pool of given Pool IDs.
    """

    logger = current_app.logger

    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Get pool info of ' + pool)

    result = get_pool_info(pool)
    return result


def get_pool_info(poolname, serialize_out=False):
    ''' returns information of the pool.
    '''
    msg = ''
    ts = time.time()
    FAILED = '"FAILED"' if serialize_out else 'FAILED'

    allpools = get_name_all_pools(
        current_app.config['FULL_BASE_LOCAL_POOLPATH'])
    if poolname in allpools:

        code, result, mes = load_HKdata([poolname], serialize_out=True)

        result = deserialize(result, int_key=True)

        burl = request.base_url
        if burl.endswith('/'):
            burl = burl[:-1]

        # Add url to tags
        dt_display = {}
        for t, clses in result['dTags'].items():
            if t == '_STID':
                continue
            cdict = {}
            for cls, sns in clses.items():
                sndict = dict((int(sn), '/'.join((burl, cls, sn)))
                              for sn in sns)
                cdict[cls] = sndict
            dt_display[t] = cdict
        display = {'Tags': dt_display}
        # add urls to urns
        ty_display = {}
        rec_u = 0
        for cl, sns in result['dTypes'].items():
            if cl == '_STID':
                continue
            snd_url = {}
            # sns is like {'currentSN': 2,
            # 'sn': {'0': {'tags': [], 'meta': [14, 2779]}, '1': ...
            rec_u += len(sns['sn'])
            for sn, snd in sns['sn'].items():
                # str makes a list to show in ome line
                base = {'tags': str(snd['tags']),
                        'meta': str(snd['meta']),
                        'url': '/'.join((burl, cl, str(sn)))
                        }
                snd_url[sn] = base
            ty_display[cl] = snd_url
        display['DataTypes'] = ty_display
        msg = 'Getting pool %s information. %s.' % (poolname, mes)
        _, count, _, _ = get_data_count(None, poolname)
        msg += '%d data items recorded. %d counted.' % (rec_u, count)
    else:
        code, result, msg = 404, FAILED, poolname + ' is not an exisiting Pool ID.'
        display = {'result': result}
    return resp(code, display, msg, ts, False)


######################################
####  {pooolid}/register PUT /unreg DELETE  ####
######################################

@ pools_api.route('/<string:pool>/', methods=['PUT'])
@ pools_api.route('/<string:pool>', methods=['PUT'])
@ auth.login_required(role='read_write')
def register(pool):
    """ Register the given pool.

    Register the pool of given Pool IDs to the global PoolManager.

    :return: response made from http code, poolurl, message
    """

    logger = current_app.logger

    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('register pool ' + pool)

    code, thepool, msg = register_pool(pool, auth.current_user())
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('PoolMgr GLB id %x' % id(PM_S._GlobalPoolList))

    res = thepool if issubclass(thepool.__class__, str) else thepool._poolurl
    return resp(code, res, msg, ts)


def register_pool(pool, usr):
    """ Register this pool to PoolManager.

    :returns: code, pool object if successful, message
    """

    logger = current_app.logger
    poolname = pool
    fullpoolpath = join(
        current_app.config['FULL_BASE_LOCAL_POOLPATH'], poolname)
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    # read_write has precedence over read_only
    makenew = usr and ('read_write' in usr.roles)
    try:
        po = PM_S.getPool(poolname=poolname, poolurl=poolurl, makenew=makenew)
        return 200, po, 'register pool ' + poolname + ' OK.'
    except (ValueError, NotImplementedError, PoolNotFoundError) as e:
        code, result, msg = excp(
            e,
            msg='Unable to register pool: ' + poolname)
        logger.error(msg)
        return code, result, msg


@ pools_api.route('/<string:pool>', methods=['DELETE'])
@ auth.login_required(role='read_write')
def unregister(pool):
    """ Unregister this pool from PoolManager.

        Check if the pool exists in server, and unregister or raise exception message to client.

    """
    logger = current_app.logger
    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('Unregister pool ' + pool)

    code, result, msg = unregister_pool(pool)
    return resp(code, result, msg, ts)


def unregister_pool(pool):
    """ Unregister this pool from PoolManager.

    Check if the pool exists in server, and unregister or raise exception message.
    :return: http code, return value, message.
    """

    poolname = pool
    current_app.logger.debug('UNREGISTER (DELETE) POOL' + poolname)
    try:
        result = PM_S.remove(poolname)
        if result == 1:
            result = '1'
            msg = 'Pool not registered or referenced: ' + poolname
            code = 200
        elif result == 0:
            result = '0'
            msg = 'Unregister pool ' + poolname + ' OK.'
            code = 200
        else:
            result = 'FAILED'
            msg = 'Unable to unregister pool: ' + poolname
            code = 409
        # current_app.logger.debug(f'{code}; {result}; {msg}')
    except Exception as e:
        code, result, msg = excp(
            e,
            msg='Unable to unregister pool: ' + poolname)
        current_app.logger.error(msg)
    checkpath.cache_clear()
    return code, result, msg


######################################
####  {pool}/hk/          ####
######################################


# @ pools_api.route('/<string:pool>/hk', methods=['GET'])
@ pools_api.route('/<string:pool>/hk/', methods=['GET'])
@ auth.login_required(role=['read_only', 'read_write'])
def hk(pool):
    """ All kinds of pool housekeeping data.

    """

    logger = current_app.logger

    ts = time.time()
    pool = pool.strip('/')
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('get HK for ' + pool)

    code, result, msg = load_HKdata([pool, 'hk'])

    return resp(code, result, msg, ts, serialize_out=True)


def load_HKdata(paths, serialize_out=True):
    """Load HKdata of a pool
    """

    poolname = paths[0]
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    # resourcetype = fullname(data)

    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.readHK(serialize_out=serialize_out)
        msg = 'OK.'
        code = 200
    except Exception as e:
        code, result, msg = excp(e, serialize_out=serialize_out)
        raise e
    return code, result, msg


################################################
####  {pool}/api/        List webAPIs       ####
################################################


@ pools_api.route('/<string:pool>/api/', methods=['GET'])
@ pools_api.route('/<string:pool>/api', methods=['GET'])
@ auth.login_required(role=['read_only', 'read_write'])
def api_info(pool):
    """ A list of names of allowed API methods.

    Returns a list of name of methods allowed to be called with web APIs on this pool.
    """

    logger = current_app.logger

    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'get allowed API methods for {pool}')

    return resp(200, WebAPI, 'OK.', ts, serialize_out=False)


######################################
#### /{pool}/wipe  PUT/  ####
######################################


@ pools_api.route('/<string:pool>/wipe', methods=['PUT'])
@ pools_api.route('/<string:pool>/wipe/', methods=['PUT'])
@ auth.login_required(role='read_write')
def wipe(pool):
    """ Removes all contents of the pool.

    requests all data in the pool be removed.
    """
    ts = time.time()
    logger = current_app.logger
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'wipe ' + pool)

    good, bad = wipe_pools([pool])
    if bad:
        code = 416
        result = 'FAILED'
        msg = 'Unable to wipe ' + pool + ' %s %s' % (str(good), str(bad))
    else:
        code = 200
        result = 0
        msg = 'Wiping %s done.' % pool

    return resp(code, result, msg, ts)

######################################
####  {pool}/hk/{kind}          ####
######################################


@ pools_api.route('/<string:pool>/hk/<string:kind>', methods=['GET'])
@ auth.login_required(role=['read_only', 'read_write'])
def hk_single(pool, kind):
    """ Returns the given kind of pool housekeeping data.
    """

    logger = current_app.logger

    ts = time.time()
    pool = pool.strip('/')
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'get {kind} HK for ' + pool)

    code, result, msg = load_single_HKdata([pool, 'hk', kind])

    return resp(code, result, msg, ts, serialize_out=True)


def load_single_HKdata(paths, serialize_out=True):
    """ Returns pool housekeeping data of the specified type: classes or urns or tags.
    """

    hkname = paths[-1]
    # paths[-2] is 'hk'
    poolname = '/'.join(paths[: -2])
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    # resourcetype = fullname(data)
    if hkname not in HKDBS:
        raise ValueError('Invalid HK type. Must be one of '+', '.join(HKDBS))
    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.readHK(hkname, serialize_out=serialize_out)
        code, msg = 200, hkname + ' HK data returned OK'
    except Exception as e:
        code, result, msg = excp(e, serialize_out=serialize_out)
    return code, result, msg

######################################
####  {pool}/count/{data_type}          ####
######################################


@ pools_api.route('/<string:pool>/count/', methods=['GET'])
@ pools_api.route('/<string:pool>/count', methods=['GET'])
@ pools_api.route('/<string:pool>/count/<string:data_type>/', methods=['GET'])
@ pools_api.route('/<string:pool>/count/<string:data_type>', methods=['GET'])
def count(pool, data_type=None):
    """ Returns the number of all types or of a given type of data in the given pool.

    Parameter

    :data_type:  (part of) dot-separated full class name of data items in pool.
    """
    logger = current_app.logger
    # return count_general(pool=pool, data_type=data_type, logger=current_app.logger)

    # def count_general(pool, data_type, logger):
    ts = time.time()
    poolname = pool.strip('/')
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'get {data_type} count for ' + pool)

    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.getCount(data_type)
        code, msg = 200, 'Recorded count of %s returned OK' % (
            data_type if data_type else 'all types')
    except Exception as e:
        code, result, msg = excp(e, serialize_out=False)

    return resp(code, result, msg, ts, serialize_out=False)


@ pools_api.route('/<string:pool>/counted/', methods=['GET'])
@ pools_api.route('/<string:pool>/counted', methods=['GET'])
@ pools_api.route('/<string:pool>/counted/<string:data_type>/', methods=['GET'])
@ pools_api.route('/<string:pool>/counted/<string:data_type>', methods=['GET'])
def counted(pool, data_type=None):
    """ Counts and Returns the number of given type of data in the given pool.

    :data_type1:  (part of) dot-separated full class name of data items in pool.
    """

    logger = current_app.logger

    ts = time.time()
    pool = pool.strip('/')
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'count {data_type} type for ' + pool)

    code, result, msg, size = get_data_count(data_type=data_type, pool_id=pool)

    return resp(code, result, msg, ts, serialize_out=False)


def get_data_count(data_type, pool_id):
    """ Return the total count for the given product type, or all types, and pool_id in the directory.

    'data_type': (part of) 'clsssname'. `None` for all types.
    'pool_id': 'pool name'

    """

    logger = current_app.logger
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('### method %s data_type %s pool %s***' %
                     (request.method, data_type, pool_id))
    res = 0
    nm = []

    path = join(current_app.config['FULL_BASE_LOCAL_POOLPATH'], pool_id)
    size = 0
    if os.path.exists(path):
        for i in os.listdir(path):
            size += os.path.getsize(join(path, i))
            if i[-1].isnumeric() and ((data_type is None) or (data_type in i)):
                res = res+1
                nm.append(i)
    else:
        return 400, 'FAILED', f'Pool {pool} not found.', 0
    s = str(nm)
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(('All types' if data_type is None else data_type) +
                     ' found ' + lls(s, 120))
    return 200, res, 'Counted %d %s files OK.' % (res, data_type if data_type else 'all types'), size


######################################
####  {pool}/api/{method}/{args} ####
######################################


@ pools_api.route('/<string:pool>/api/<string:method_args>/', methods=['GET', 'POST'])
@ pools_api.route('/<string:pool>/api/<string:method_args>', methods=['GET', 'POST'])
@ auth.login_required(role='read_write')
def api(pool, method_args):
    """ Call api mathods on the running pool and returns the result.

    """

    logger = current_app.logger

    ts = time.time()
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('get API for %s, %s(%d args).' %
                     (pool, method_args[:6], len(method_args)))
    if request.method == 'POST':
        # long args are sent with POST
        if request.data is None:
            result, msg = '"FAILED"', 'No Request data for command '+request.method
            code = 400
            return resp(code, result, msg, ts, serialize_out=True)
        data = request.get_data(as_text=True)
        paths = [pool, 'api', method_args, data]
    else:
        paths = [pool, 'api', method_args]
    lp0 = len(paths)

    posted = request.method == 'POST'
    code, result, msg = call_pool_Api(
        paths, serialize_out=False, posted=posted)

    return resp(code, result, msg, ts, serialize_out=False)


def call_pool_Api(paths, serialize_out=False, posted=False):
    """ Call api mathods on the running pool and returns the result.

    return: value if args is pool property; execution result if method.
    """

    FAILED = '"FAILED"' if serialize_out else 'FAILED'
    logger = current_app.logger
    ts = time.time()

    if posted:
        code = 200
        try:
            m_args, kwds = tuple(deserialize(paths[3]))
        except ValueError as e:
            code = 422
        m_args.insert(0, paths[2])
    else:
        args, kwds = [], {}

        # the unquoted args. may have ',' in strings
        # quoted_m_args = paths[ind_meth+1]

        # from the unquoted url extract the fist path segment.
        quoted_m_args = request.url.split(
            paths[0] + '/' + paths[1] + '/')[1].strip('/')
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug('get API : %s' % lls(quoted_m_args, 1000))
        # get command positional arguments and keyword arguments
        code, m_args, kwds = deserialize_args(
            quoted_m_args, serialize_out=serialize_out)

    if code != 200:
        result, msg = m_args, kwds
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(f'RT{code} {msg}')
        return 0, resp(422, result, msg, ts, serialize_out=False), 0

    method = m_args[0]

    if method not in WebAPI:
        code = 400
        msg = 'Unknown web API method: %s.' % method
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(f'RT{code} {msg}')
        return 0, resp(code, FAILED,
                       msg,
                       ts, serialize_out=False), 0
    args = m_args[1:] if len(m_args) > 1 else []
    kwdsexpr = [str(k)+'='+str(v) for k, v in kwds.items()]
    msg = '%s(%s)' % (method, ', '.join(
        chain(((getattr(x, 'where', str(x)))[:100] for x in args), kwdsexpr)))
    if logger.isEnabledFor(logging_DEBUG):
        logger.debug('WebAPI ' + lls(msg, 300) +
                     's %x' % id(PM_S._GlobalPoolList))

    if method == 'isEmpty':
        print('********', list(PM_S._GlobalPoolList))

    poolname = paths[0]
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    if not PM_S.isLoaded(poolname):
        msg = 'Pool not found or not registered: ' + poolname
        if method in ('removeAll'):
            result = 'OK'
            code = 200
            if logger.isEnabledFor(logging_DEBUG):
                logger.debug(f'RT{code} {msg}')
            return 0, resp(code, result, msg, ts, serialize_out=False), 0
        else:
            result = FAILED
            if logger.isEnabledFor(logging_ERROR):
                logger.error(msg)
            code = 404
            if logger.isEnabledFor(logging_DEBUG):
                logger.debug(f'RT{code} {msg}')
            return 0, resp(code, result, msg, ts, serialize_out=False), 0

    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        res = getattr(poolobj, method)(*args, **kwds)
        result = res
        msg = msg + ' OK.'
        code = 200
    except Exception as e:
        code, result, msg = excp(e, 422, serialize_out=serialize_out)
        logger.error(msg)

    if logger.isEnabledFor(logging_DEBUG):
        logger.debug(f'RT{code} {msg}')
    return 0, resp(code, result, msg, ts, serialize_out=False), 0



__version_info__ = (1, 28, 0)
__version__ = '.'.join(map(str, __version_info__))
# 1.28.0 csdb overaul and asynhttp. bug fixes,
# 1.27.0 list input to `remove` localpool and subclasses. async io for HttpClientPool
# 1.26.2 Implement namespace-based getConfig(). Fix unregister bug in PS.remove()
# 1.26.1 simplify user config for httppool.
# 1.26.0 relate the 104 Connection Reset error to auto redirect to trailing '/' url. Add '/' to endpoints. add "/pools" to show pool info.
# 1.25.4 threaded test code. local httppool read/write prod 20-30ms
# 1.25.3 refactor query code
# 1.25.2 server remote debug
# 1.25.1 externalize docker tests.
# 1.25.0 getConfig and configuration improvement
# 1.24.4 requets timeout; logger_level_extras; single thread and python3.8 in wsgi; other improvement to get self-test pass for server docker.
# 1.24.3 improved config.py for docker, RW_USER, missing templates.
# 1.24.2 session works and all tests with live (fore/background) pools, w or w/o session. py3.9 in readthdoc.yml.'
# 1.24.1 intersphinx enabled for docs.
# 1.24.0 session for httppool, testing wih live mock server
# 1.23.1 greatly simplfies MetaData.toString with new Python-tabulate.
# 1.23.0 session and templates for httppool; extra fixed.
# 1.22.3 History docs and working on server
# 1.22.2 improve yaml2python parent sort; implement history tracing; Documents layout fix; Parameter takes anything as value unchecked.




from fdi.dataset.arraydataset import ArrayDataset

from fdi.pal.mempool import MemPool
from fdi.pal.poolmanager import PoolManager, DEFAULT_MEM_POOL
from fdi.utils.common import trbk, fullname
from fdi.pal.context import Context, MapContext, RefContainer
from fdi.pal.productref import ProductRef
from fdi.pal.productstorage import ProductStorage
from fdi.pal.urn import Urn, parseUrn, parse_poolurl, makeUrn, UrnUtils
from fdi.pal.productpool import ProductPool
from fdi.pal.localpool import LocalPool
from fdi.pal.httpclientpool import HttpClientPool
from fdi.pal.context import Context
from fdi.pal.query import AbstractQuery, MetaQuery
from fdi.dataset.deserialize import deserialize
from fdi.dataset.product import Product
from fdi.dataset.baseproduct import BaseProduct
from fdi.dataset.eq import deepcmp
from fdi.dataset.classes import Classes
from fdi.dataset.metadata import MetaData, Parameter
from fdi.dataset.finetime import FineTime1
from fdi.dataset.testproducts import TP
from fdi.utils.checkjson import checkjson
from fdi.pns.fdi_requests import save_to_server, read_from_server

# from flask import request as
import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import ConnectionError
from urllib3.exceptions import NewConnectionError
import itertools
import random
import timeit
import pytest
import copy
import traceback
from pprint import pprint
import json
import shutil
import getpass
import os
import weakref
import gc
from os import path as op
import glob


import sys

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
else:
    PY3 = False


Test_Pool_Name = __name__.replace('.', '_')
defaultpoolPath = '/tmp/fditest'

if __name__ == '__main__' and __package__ == 'tests':
    # run by python -m tests.test_dataset

    Test_Pool_Name = 'test_pal'

    pass

else:
    # run by pytest

    # This is to be able to test w/ or w/o installing the package
    # https://docs.python-guide.org/writing/structure/
    from pycontext import fdi

    from logdict import logdict
    import logging
    import logging.config
    # create logger
    logging.config.dictConfig(logdict)
    logger = logging.getLogger()
    logger.debug('logging level %d' %
                 (logger.getEffectiveLevel()))
    logging.getLogger("filelock").setLevel(logging.WARNING)


def checkgeneral(v):
    # can always add attributes
    t = 'random'
    v.testattr = t
    assert v.testattr == t
    try:
        m = v.notexists
    except AttributeError as e:
        assert str(e).split()[-1] == "'notexists'", traceback.print_exc()
    except:
        traceback.print_exc()
        assert false


def test_UrnUtils():
    prd = Product(description='pal test')
    a1 = 'file'      # scheme
    a2 = '/e:'            # place
    b1, b2 = '/tmp/foo', 'name'
    a3 = b1 + '/' + b2               # /tmp/foo/name
    a4 = fullname(prd)           # fdi.dataset.Product
    a5 = 43
    s = a1 + '://' + a2          # file:///e:
    poolurl = s + a3                   # file:///e:/tmp/foo/name
    r = a4 + ':' + str(a5)       # fdi.dataset.Product:43
    rp = a4 + '_' + str(a5)      # fdi.dataset.Product_43
    urn = 'urn:' + b2 + ':' + r  # urn:name:fdi.dataset.Product:43
    urn1 = 'urn:' + b2 + ':' + a4+':'+str(a5-1)
    # utils
    assert parseUrn(None) == (None, None, None)
    assert parseUrn(urn) == (b2, a4, a5)
    assert parseUrn([urn, urn]) == (b2, a4, [a5, a5])
    assert parseUrn(urn) == (b2, a4, a5)
    assert parseUrn([urn, f'urn:{b2+"a"}:{a4+"b"}:{a5+3}']) == \
        ([b2, b2+"a"], [a4, a4+"b"], [a5, a5+3])
    assert UrnUtils.isUrn(urn)
    assert UrnUtils.getProductId(urn) == a5
    assert UrnUtils.getPoolId(urn) == b2
    assert UrnUtils.getLater(urn1, urn) == urn
    assert UrnUtils.getClassName(urn) == a4
    assert UrnUtils.getClass(urn).__name__ == a4.split('.')[-1]
    assert UrnUtils.extractRecordIDs([urn, urn1]) == [a5, a5-1]

    # assert UrnUtils.getPool(urn,pools)
    # assert UrnUtils.containsUrn(urn, pool)
    assert UrnUtils.checkUrn(urn)
    with pytest.raises(ValueError):
        UrnUtils.checkUrn(urn+'r')

    # poolurl
    poolpath, scheme, place, poolname, un, pw = parse_poolurl(
        poolurl, poolhint=b2)
    assert poolpath == a2 + b1
    assert scheme == a1
    assert place == ''
    assert poolname == b2
    # assert (un, pw) == ('foo', 'bar')
    # implicit pool is the last segment
    poolpath, scheme, place, poolname, un, pw = parse_poolurl(
        'file:///c:/tmp/mypool/v3/')
    assert poolpath == '/c:/tmp/mypool'
    assert scheme == 'file'
    assert place == ''
    assert poolname == 'v3'
    # explicit poolname. the first distinctive substring
    poolpath, scheme, place, poolname, un, pw = parse_poolurl(
        'file:///c:/tmp/mypool/v3/', 'my')
    assert poolpath == '/c:/tmp'
    assert scheme == 'file'
    assert place == ''
    assert poolname == 'mypool/v3'

    # https scheme. pool parameter is given in a urn
    poolpath, scheme, place, poolname, un, pw = parse_poolurl(
        'https://127.0.0.1:5000/v3/mypool/v2', 'urn:mypool/v2:foo.KProduct:43')
    assert poolpath == '/v3'
    assert scheme == 'https'
    assert place == '127.0.0.1:5000'
    assert poolname == 'mypool/v2'


def test_Urn():
    prd = BaseProduct(description='pal test')
    a1 = 'file'      # scheme
    a2 = '/e:'            # place
    b1, b2 = '/tmp/foo', 'name'
    a3 = b1 + '/' + b2               # /tmp/foo/name
    a4 = fullname(prd)           # fdi.dataset.BaseProduct
    a5 = 43
    s = a1 + '://' + a2          # file:///e:
    poolurl = s + a3                   # file:///e:/tmp/foo/name
    r = a4 + ':' + str(a5)       # fdi.dataset.BaseProduct:43
    rp = a4 + '_' + str(a5)      # fdi.dataset.BaseProduct_43
    urn = 'urn:' + b2 + ':' + r  # urn:name:fdi.dataset.BaseProduct:43
    urn1 = 'urn:' + b2 + ':' + a4+':'+str(a5-1)
    # constructor
    # urn only
    v = Urn(urn=urn)
    assert v.getPoolId() == b2
    assert v.getUrnWithoutPoolId() == r
    assert v.getIndex() == a5
    assert v.getUrn() == urn
    assert v.getScheme() is None
    assert v.getPlace() is None
    assert v.getPoolpath() is None

    # urn with poolurl
    v = Urn(urn=urn, poolurl=poolurl)
    assert v.getPoolId() == b2  #
    assert v.getUrnWithoutPoolId() == r
    assert v.getIndex() == a5
    assert v.getUrn() == urn
    assert v.getScheme() == a1
    assert v.getPlace() == ''
    assert v.getPoolpath() == a2 + b1
    # urn with components
    v = Urn(cls=prd.__class__, poolname=b2, index=a5)
    assert v.getPoolId() == b2
    assert v.getUrnWithoutPoolId() == r
    assert v.getIndex() == a5
    assert v.getUrn() == urn
    assert v.getScheme() is None
    assert v.getPlace() is None
    assert v.getPoolpath() is None
    # no urn then other args must all be given
    with pytest.raises(ValueError):
        v = Urn(cls=prd.__class__, poolname=b2)

    # no-arg constructor
    v = Urn()
    v.urn = urn
    assert v.getPoolId() == b2
    assert v.getUrnWithoutPoolId() == r
    assert v.getIndex() == a5
    assert v.getUrn() == urn
    assert v.getScheme() is None
    assert v.getPlace() is None
    assert v.getPoolpath() is None
    v = Urn()
    v.setUrn(urn=urn, poolurl=poolurl)
    assert v.getPoolId() == b2
    assert v.getUrnWithoutPoolId() == r
    assert v.getIndex() == a5
    assert v.getUrn() == urn
    assert v.getScheme() == a1
    assert v.getPlace() == ''
    assert v.getPoolpath() == a2 + b1

    # access
    assert v.getUrn() == v.urn
    assert v.getPool() == v.pool
    assert v.getTypeName() == a4
    assert v.getPlace() == v.place

    checkjson(v)


def transpath(direc, poolpath):
    """ direc must have a leading / if base_local_poolpath is defined in config """

    return poolpath+'/'+direc


def rmlocal(d):
    if op.exists(d):
        try:
            # print(os.stat(d))
            shutil.rmtree(d)
        except Exception as e:
            print(str(e) + ' ' + trbk(e))
            raise
        assert not op.exists(d)


def cleanup(poolurl=None, poolname=None):
    """ remove pool from disk and memory"""

    if poolurl or poolname:
        name_url = [(poolname, poolurl)]
    else:
        name_url = []
        for pn, pool in PoolManager.getMap().items():
            name_url.append((pn, pool._poolurl))
    for pname, purl in name_url:
        direc, schm, place, pn, un, pw = parse_poolurl(purl, pname)
        if schm in ['file', 'server']:
            d = direc + '/' + pn
            rmlocal(d)
        elif schm == 'mem':
            if PoolManager.isLoaded(DEFAULT_MEM_POOL):
                PoolManager.getPool(DEFAULT_MEM_POOL).removeAll()
        elif schm in ['http', 'https']:
            pass
        elif schm in ['csdb']:
            pass
        else:
            assert False
        # correct way
        try:
            p = PoolManager.getPool(pname, poolurl)
            p.removeAll()
        except (RuntimeError, ValueError):
            pass

        PoolManager.remove(pname)
        try:
            del p
        except NameError:
            pass
        gc.collect()


def test_PoolManager():
    defaultpoolName = Test_Pool_Name
    defaultpoolUrl = 'file://' + defaultpoolPath + '/' + defaultpoolName
    cleanup(defaultpoolUrl, defaultpoolName)
    # class methods

    assert PoolManager.size() == 0
    # This creates a pool and returns it if the pool of given name does not exist
    pool = PoolManager.getPool(defaultpoolName, defaultpoolUrl)
    assert PoolManager.size() == 1
    assert defaultpoolName in PoolManager.getMap()
    # print('GlobalPoolList#: ' + str(id(pm.getMap())) + str(pm))
    PoolManager.remove('not_exists', ignore_error=True)
    with pytest.raises(KeyError):
        PoolManager.remove('not_exists')
    PoolManager.removeAll()
    assert PoolManager.size() == 0
    assert weakref.getweakrefcount(pool) == 0
    # print(weakref.getweakrefs(pool), id(pool), 'mmmm pool')
    del pool

    # initiate
    pm = PoolManager()
    assert len(pm) == 0
    p1 = pm.getPool(defaultpoolName, defaultpoolUrl)
    # print(weakref.getweakrefs(p1), id(p1), 'mmmm p1')
    for k, v in pm.items():
        assert isinstance(v, ProductPool)
    assert defaultpoolName in pm
    PG = PoolManager._GlobalPoolList
    assert PoolManager.isLoaded(defaultpoolName) == 1
    del v
    assert PoolManager.isLoaded(defaultpoolName) == 1
    # print(weakref.getweakrefs(PG[defaultpoolName]), id(
    #    PG[defaultpoolName]), 'mmm GL')
    pr = weakref.ref(p1)
    assert pr() == p1
    assert weakref.getweakrefcount(pr()) == 2
    # print(weakref.getweakrefs(PG[defaultpoolName]),
    #      id(PG[defaultpoolName]), 'mmm+r')
    del p1
    assert pr()
    # print(weakref.getweakrefs(PG[defaultpoolName]),
    #      id(PG[defaultpoolName]), 'mmmm del p1')
    assert weakref.getweakrefcount(PG[defaultpoolName]) == 2
    gc.collect()
    # print(weakref.getweakrefs(PG[defaultpoolName]),
    #      id(PG[defaultpoolName]), 'mmmm gc')
    assert weakref.getweakrefcount(PG[defaultpoolName]) == 2  # why ?
    del pr
    # print(weakref.getweakrefs(PG[defaultpoolName]),
    #      id(PG[defaultpoolName]), 'mmmm del r')
    assert weakref.getweakrefcount(PG[defaultpoolName]) == 1
    assert pm.remove(defaultpoolName) == 0

    # http pool gets registered
    try:
        with pytest.raises(ConnectionError):
            ph = pm.getPool(poolurl='http://h.edu/foo')
    except NewConnectionError:
        pass
    assert not PoolManager.isLoaded('foo')
    with pytest.raises(KeyError):
        assert PoolManager.remove('foo') == 1


def checkdbcount(expected_cnt, poolurl, prodname, currentSN, usrpsw, *args, csdb=None, **kwds):
    """ init_count files in pool and entries in class db.

    expected_cnt, currentSN: expected number of prods and currentSN in pool for products named prodname
    """
    poolpath, schm, place, poolname, un, pw = parse_poolurl(poolurl)
    scheme = schm.lower()
    if scheme in ['file', 'server']:
        path = op.join(poolpath, poolname)
        assert sum(1 for x in glob.glob(
            op.join(path, prodname + '*[0-9]'))) == expected_cnt
        cp = op.join(path, 'dTypes.jsn')
        if op.exists(cp) or expected_cnt != 0:
            with open(cp, 'r') as fp:
                js = fp.read()
            cread = deserialize(js, int_key=True)
            if currentSN is None:
                assert cread[prodname]['currentSN'] == currentSN
                # number of items is expected_cnt
            assert len(cread[prodname]['sn']) == expected_cnt
    elif scheme == 'mem':
        mpool = PoolManager.getPool(poolurl=poolurl).getPoolSpace()
        if mpool is None or len(mpool) == 0:
            # wiped
            assert expected_cnt == 0
            assert currentSN is None
            return
        ns = [n for n in mpool if prodname in n]
        assert len(ns) == expected_cnt, len(ns)
        if currentSN is None:
            assert mpool['dTypes'][prodname]['currentSN'] == currentSN
        # for this class there are  how many prods
        assert len(mpool['dTypes'][prodname]['sn']) == expected_cnt
    elif scheme in ['http', 'https']:
        auth = HTTPBasicAuth(*usrpsw)
        # count
        cpath = poolname + '/' + 'count/' + prodname
        api_baseurl = scheme + '://' + place + poolpath + '/'
        url = api_baseurl + cpath
        x = requests.get(url, auth=auth)
        assert 'Recorded' in x.text
        count = int(x.json()['result'])
        assert count == expected_cnt
        # counting files. Append a '/'
        cpath = poolname + '/' + 'counted/' + prodname + '/'
        api_baseurl = scheme + '://' + place + poolpath + '/'
        url = api_baseurl + cpath
        x = requests.get(url, auth=auth)
        assert 'Counted' in x.text
        count = int(x.json()['result'])
        assert count == expected_cnt

        # count all
        cpath = poolname + '/' + 'count'
        api_baseurl = scheme + '://' + place + poolpath + '/'
        url = api_baseurl + cpath
        x = requests.get(url, auth=auth)
        assert 'Recorded' in x.text
        count = int(x.json()['result'])
        assert count >= expected_cnt

        # count all with couting
        cpath = poolname + '/' + 'counted' + '/'
        api_baseurl = scheme + '://' + place + poolpath + '/'
        url = api_baseurl + cpath
        x = requests.get(url, auth=auth)
        assert 'Counted' in x.text
        count1 = int(x.json()['result'])
        assert count1 >= expected_cnt

        assert count == count1

        # sn
        if currentSN is None:
            return
        spath = poolname + '/' + 'hk/dTypes'
        url = api_baseurl + spath
        x = requests.get(url, auth=auth)
        csn = x.json()['result'][prodname]['currentSN']
        assert csn == currentSN

    elif scheme in ['csdb']:
        test_pool = PoolManager.getPool(poolname)
        # a, b = csdb
        # assert test_pool == csdb
        pinfo = test_pool.getPoolInfo()
        count = test_pool.getCount(prodname)
        assert count == expected_cnt
        if currentSN is None:
            return
        for cl in pinfo[poolname]['_classes']:
            if cl['productTypeName'] == prodname:
                sn = cl['sn']
                assert cl['currentSn'] == currentSN
                break
    else:
        assert False, 'bad pool scheme'


def test_ProductRef():
    defaultpoolName = Test_Pool_Name
    defaultpoolUrl = 'file://' + defaultpoolPath + '/' + defaultpoolName
    cleanup(defaultpoolUrl, defaultpoolName)
    prd = Product()
    a1 = 'file'
    a2 = defaultpoolPath
    a3 = defaultpoolName
    a4 = fullname(prd)
    a5 = 0
    s = a1 + '://' + a2   # file:///tmp
    p = s + '/' + a3  # a pool URL
    r = a4 + ':' + str(a5)  # a resource
    u = 'urn:' + a3 + ':' + r    # a URN
    cleanup(p, a3)

    cleanup('mem:///' + DEFAULT_MEM_POOL, DEFAULT_MEM_POOL)

    # in memory
    # A productref created from a single product will result in a memory pool urn, and the metadata won't be loaded.
    v = ProductRef(prd)
    # only one prod in memory pool
    checkdbcount(1, 'mem:///' + DEFAULT_MEM_POOL, a4, 0, '')
    assert v.urn == 'urn:' + DEFAULT_MEM_POOL + ':' + a4 + ':' + str(0)
    assert v.meta is None
    assert v.product == prd
    # run again will get a different urn from the new instance in the pool
    r = ProductRef(prd)
    assert v != r

    # construction

    ps = ProductStorage(a3, poolurl=p)
    prd = Product()
    rfps = ps.save(prd)
    pr = ProductRef(urn=rfps.urnobj, poolname=a3)
    assert rfps == pr
    assert rfps.getMeta() == pr.getMeta()
    uobj = Urn(urn=u)
    assert pr.urnobj == uobj
    # load given metadata
    met = MetaData()
    met['foo'] = Parameter('bar')
    prm = ProductRef(urn=u, meta=met)
    assert prm.meta['foo'].value == 'bar'
    # This does not obtain metadata
    pr = ProductRef(urn=rfps.urnobj)
    assert rfps == pr
    assert rfps.getMeta() != pr.getMeta()
    assert pr.urnobj == uobj
    assert pr.getPoolname() == a3
    assert rfps.getPoolname() is not None
    # load from a storage.
    pr = ps.load(u)
    assert rfps == pr
    assert rfps.getMeta() == pr.getMeta()
    assert pr.getPoolname() == rfps.getPoolname()

    # parent
    # nominal ops
    b1 = Product(description='abc')
    b2 = MapContext(description='3c273')
    pr.addParent(b1)
    pr.addParent(b2)
    assert b1 in list(pr.parents)
    assert b2 in list(pr.parents)

    pr.removeParent(b1)
    assert b1 not in list(pr.parents)
    # access
    assert pr.urnobj.getTypeName() == a4
    assert pr.urnobj.getIndex() == a5
    # this is tested in ProdStorage
    # assert pr.product == p

    checkjson(pr)


def test_ProductStorage_init():
    defaultpoolname = Test_Pool_Name
    poolpath = '/tmp/fditest'
    defaultpoolurl = 'file://' + poolpath + '/' + defaultpoolname
    cleanup(defaultpoolurl, defaultpoolname)
    newpoolname = 'new_' + Test_Pool_Name
    newpoolurl = 'file://' + poolpath + '/' + newpoolname
    cleanup(newpoolurl, newpoolname)

    # Constructor
    # no default pool
    ps = ProductStorage()
    with pytest.raises(IndexError):
        p1 = ps.getPools()[0]
    # with poolurl the pool is constructed
    ps = ProductStorage(poolurl=defaultpoolurl)
    # There is no product
    assert ps.isEmpty()
    # get the pool object
    pspool = ps.getPool(defaultpoolname)
    assert len(pspool.getProductClasses()) == 0
    # check syntax: construct a storage with a poolobj
    pool = LocalPool(poolurl=defaultpoolurl)
    ps2 = ProductStorage(pool)
    assert ps.getPools() == ps2.getPools()
    # wrong poolname
    with pytest.raises(TypeError):
        psbad = ProductStorage(defaultpoolurl)

    # register pool
    # with a storage that already has a pool

    ps2.register(newpoolname, newpoolurl)
    assert op.exists(transpath(newpoolname, poolpath))
    assert len(ps2.getPools()) == 2
    assert ps2.getPools()[1] == newpoolname

    # multiple storages pointing to the same pool will get exception
    # with pytest.raises(TypeError):
    ps2 = ProductStorage()


def getCurrSnCount(csdb_c, prodname):
    pinfo = csdb_c if issubclass(csdb_c.__class__, dict) else \
        csdb_c.getPoolInfo()
    init_sn, init_count = 0, 0
    for pool in pinfo.values():
        poolclass = pool['_classes']
        if prodname in poolclass:
            init_sn = poolclass[prodname]['currentSn']
            # XXX was the issue xaused " - 1" fixed?
            init_count = len(poolclass[prodname]['sn'])
            break
    return init_sn, init_count, pinfo


def check_prodStorage_func_for_pool(thepoolname, thepoolurl, *args):
    ps = ProductStorage(poolurl=thepoolurl)
    p1 = ps.getPools()[0]
    # get the pool object
    pspool = ps.getPool(p1)
    x = Product(description="This is my product example",
                instrument="MyFavourite", modelName="Flight")
#    y = MapContext(description='Keep it all un context.')
    pcq = fullname(x)
    mcq = fullname(MapContext)

    is_csdb = thepoolurl.startswith('csdb://')
    if is_csdb:
        csdb = PoolManager.getPool(thepoolname)
        init_sn, init_count, pinfo = getCurrSnCount(csdb, pcq)
        init_sn_m, init_count_m, _ = getCurrSnCount(pinfo, mcq)
        logger.debug('1...', init_count, init_sn, init_count_m, init_sn_m)
    # save
    ref = ps.save(x)
    s0, s1 = tuple(ref.urn.rsplit(':', 1))
    assert s0 == 'urn:' + thepoolname + ':' + pcq
    # not saving this one
    # ref_m = ps.save(y)
    init_count, init_count_m = 0, 0
    if is_csdb:
        __import__("pdb").set_trace()

        currentsn2, count2, pinfo2 = getCurrSnCount(csdb, pcq)
        # m is unchanged
        currentsn_m2, count_m2, pinfo_m2 = getCurrSnCount(pinfo2, mcq)
        logger.debug('2...', count2, currentsn2, count_m2, currentsn_m2)

        assert count2 == init_count + 1
        assert currentsn2 == init_sn + 1
        assert init_count_m == init_count_m
        assert currentsn_m2 == init_sn_m

        # XXX this 1 is due to a bug in csdb implementation
        init_sn += 1
        init_sn_m += 1
    else:
        init_sn = 0
        init_count = 0
        init_sn_m = 0
        init_count_m = 0

    # ps has 1 prod
    assert init_sn == int(s1)
    checkdbcount(init_count+1, thepoolurl, pcq, init_sn + 0, *args)
    # checkdbcount(init_count_m+0, thepoolurl, mcq, init_sn_m + 0, *args)

    # save more
    # one by one
    q = 3
    x2, ref2 = [], []

    for d in range(q):
        # The 1st one is a mapcontext, rest product
        tmp = Product(description='x' + str(d)
                      ) if d > 0 else MapContext(description='x0')
        x2.append(tmp)
        ref2.append(ps.save(tmp, tag='t' + str(d)))

    checkdbcount(init_count + q, thepoolurl, pcq,
                 init_sn + q - 1, *args)
    checkdbcount(init_count_m+1, thepoolurl, mcq, init_sn_m+0, *args)
    # save many in one go
    m, x3 = 2, []
    n = q + m
    for d in range(q, n):
        tmp = Product(description='x' + str(d))
        x3.append(tmp)
    ref2 += ps.save(x3, tag='all-tm')  # ps has n+1 prods
    x2 += x3  # there are n prods in x2
    # check refs
    assert len(ref2) == n
    checkdbcount(init_count+n, thepoolurl, pcq, init_sn + n - 1, *args)
    checkdbcount(init_count_m+1, thepoolurl, mcq, init_sn_m+0, *args)

    # tags
    # XXX fix getUrn() and getTag() when no arg   is given
    if not is_csdb:
        ts = ps.getAllTags()
        assert len(ts) == q + 1
        ts = ps.getTags(ref2[0].urn)
        assert len(ts) == 1
        assert ts[0] == 't0'
        u = ps.getUrnFromTag('all-tm')
        assert len(u) == m
        assert u[0] == ref2[q].urn

    # access resource
    checkdbcount(init_count+n, thepoolurl, pcq, init_sn+n-1, *args)
    checkdbcount(init_count_m+1, thepoolurl, mcq, init_sn_m+0, *args)
    # get ref from urn
    pref = ps.load(ref2[n - 2].urn)
    assert pref == ref2[n - 2]
    # actual product
    # print(pref._product)
    assert pref.product == x2[n - 2]
    # from tags

    # removal by reference urn
    # print(ref2[n - 2].urn)
    ps.remove(ref2[n - 2].urn)
    # files are less
    # DB shows less in record
    # current serial number not changed
    # number of items decreased by 1
    checkdbcount(init_count + n - 1, thepoolurl,
                 pcq, init_sn+n-1, *args)
    checkdbcount(init_count_m+1, thepoolurl, mcq, init_sn_m+0, *args)

    # check isEmpty

    # report lru_cache info
    print('***HTTPClient Pool cache*** %s ' % str(pspool.getCacheInfo()))

    # clean up a pool
    ps.wipePool()
    checkdbcount(0, thepoolurl, pcq, None, *args)
    assert ps.getPool(thepoolname).isEmpty()


def test_ProdStorage_func_local_mem():
    # local pool
    thepoolname = Test_Pool_Name
    thepoolpath = '/tmp/fditest'
    thepoolurl = 'file://' + thepoolpath + '/' + thepoolname
    cleanup(thepoolurl, thepoolname)
    check_prodStorage_func_for_pool(thepoolname, thepoolurl, None)

    # mempool
    thepoolname = DEFAULT_MEM_POOL
    thepoolpath = '/'
    thepoolurl = 'mem://' + thepoolpath + thepoolname

    cleanup(thepoolurl, thepoolname)
    check_prodStorage_func_for_pool(thepoolname, thepoolurl, None)


def test_ProdStorage_func_http(server, userpass):

    aburl, hdr = server
    # httpclientpool
    thepoolname = Test_Pool_Name
    thepoolurl = aburl + '/' + thepoolname

    cleanup(thepoolurl, thepoolname)
    # First test registering with local pstor will also register on server
    pool = HttpClientPool(poolurl=thepoolurl)

    # register
    pstore = ProductStorage(pool=pool)

    assert pool.isEmpty()
    pstore.unregister(thepoolname)
    # not exist
    with pytest.raises(RuntimeError):
        assert pool.isEmpty()
    check_prodStorage_func_for_pool(thepoolname, thepoolurl, userpass)


def test_ProdStorage_func_server():
    # httppool , the http server-side pool
    thepoolname = 'server'+Test_Pool_Name
    thepoolurl = 'server://' + '/tmp/fditest' + '/' + thepoolname

    cleanup(thepoolurl, thepoolname)
    check_prodStorage_func_for_pool(thepoolname, thepoolurl, None)


def test_ProdStorage_func_csdb(csdb):
    thepool, thepoolurl, pstore = csdb  # csdb:///csdb_test_pool
    thepoolname = thepool.poolname

    cleanup(thepoolurl, thepoolname)
    check_prodStorage_func_for_pool(thepoolname, thepoolurl, None)


def test_LocalPool():
    thepoolname = 'localpool_' + Test_Pool_Name
    thepoolpath = '/tmp/fditest'
    thepoolurl = 'file://' + thepoolpath + '/' + thepoolname

    cleanup(thepoolurl, thepoolname)
    ps = ProductStorage(thepoolname, thepoolurl)
    pname = ps.getPools()[0]
    # get the pool object
    pspool = ps.getPool(pname)

    x = Product(description="This is my product example",
                instrument="MyFavourite", modelName="Flight")
    pcq = fullname(x)
    # save
    ref = ps.save(x, tag='ttag')

    # read HK
    # copy default pool data in memory
    p1 = pspool
    # rename the pool
    cpn = thepoolname + '_copy'
    cpu = thepoolurl + '_copy'
    pcp = transpath(cpn, thepoolpath)
    if op.exists(pcp):
        shutil.rmtree(pcp)
    # make a copy of the old pool on disk
    shutil.copytree(transpath(thepoolname, thepoolpath), pcp)
    ps2 = ProductStorage(pool=cpn, poolurl=cpu)
    # two ProdStorage instances have the same DB
    p2 = ps2.getPool(ps2.getPools()[0])
    # assert deepcmp(p1._urns, p2._urns) is None
    # assert deepcmp(p1._tags, p2._tags) is None
    # assert deepcmp(p1._classes, p2._classes) is None
    assert deepcmp(p1._dTypes, p2._dTypes) is None
    assert deepcmp(p1._dTags, p2._dTags) is None

    # remove till empty
    ps2.save(x, tag='ttag')
    urns = ps2.getUrnFromTag('ttag')
    assert len(urns) == 2
    for u in urns:
        p2.remove(u)
    assert len(ps2.getUrnFromTag('ttag')) == 0

    # remove multiple sn
    ps2.save(x, tag='mtag')
    ps2.save(x, tag='mtag')
    urns = ps2.getUrnFromTag('mtag')
    pool, dtype, sns = parseUrn(urns)
    assert pool == cpn
    assert dtype == 'fdi.dataset.product.Product'
    assert len(sns) == 2
    p2.remove(resourcetype=dtype, index=sns)
    assert len(ps2.getUrnFromTag('mtag')) == 0

    # remove multiple sn ASYNCHRONOUSLY
    ps2.save(x, tag='mtag')
    ps2.save(x, tag='mtag')
    urns = ps2.getUrnFromTag('mtag')
    pool, dtype, sns = parseUrn(urns)
    assert pool == cpn
    assert dtype == 'fdi.dataset.product.Product'
    assert len(sns) == 2
    p2.remove(resourcetype=dtype, index=sns, asyn=True)
    assert len(ps2.getUrnFromTag('mtag')) == 0

    backup_restore(ps)


def backup_restore(ps):
    p1 = ps.getPool(ps.getPools()[0])
    hk1 = p1.readHK()
    cpn = p1._poolname + '_2'
    cpu = p1._poolurl + '_2'
    pstore = ProductStorage(pool=cpn, poolurl=cpu)
    # backup
    # the new pool is made empty
    pstore.wipePool()
    # register
    pstore = ProductStorage(pool=cpn, poolurl=cpu)
    # save something to the new pool
    x = Product(description="This is my product 2")
    ref = pstore.save(x, tag='i think')
    ref2 = pstore.save(x, tag='i think')
    assert ref != ref2
    # two pools are different
    p2 = pstore.getPool(pstore.getPools()[0])
    assert deepcmp(hk1, p2.readHK()) is not None

    # make a backup tarfile
    tar = p1.backup()
    os.makedirs('/tmp/fditest', exist_ok=True)
    with open('/tmp/fditest/bk.tar', 'wb') as f:
        f.write(tar)
    with open('/tmp/fditest/bk.tar', 'rb') as f:
        tar2 = f.read()
    assert tar == tar2
    # restore
    lst = p2.restore(tar2)

    # two pools are the same
    assert deepcmp(hk1, p2.readHK()) is None


def mkStorage(thepoolname, thepoolurl, pstore=None):
    """ returns pool object and productStorage """

    cleanup(thepoolurl, thepoolname)
    if not pstore:
        pstore = ProductStorage(thepoolname, thepoolurl)
    thepoolpath, tsc, tpl, pn, un, pw = parse_poolurl(thepoolurl, thepoolname)
    if tsc in ['file', 'server']:
        assert op.exists(transpath(thepoolname, thepoolpath))
    assert len(pstore.getPools()) == 1
    assert pstore.getPools()[0] == thepoolname
    thepool = PoolManager.getMap()[thepoolname]
    assert thepool.getScheme() == tsc
    assert thepool.isEmpty()
    return thepool, pstore


def doquery(poolpath, newpoolpath):
    # creation
    a1 = MapContext
    a2 = 'p'
    a3 = 'p.description == "mc"'
    a4 = False
    q = AbstractQuery(product=a1, variable=a2, where=a3, allVersions=a4)
    assert q.getType() == a1
    assert q.getVariable() == a2
    assert q.getWhere() == a3
    assert q.retrieveAllVersions() == a4

    a1 = TP
    a2 = 'm'
    a3 = 'm["description"].value == "pr"'
    a4 = False
    q = MetaQuery(product=a1, where=a3, allVersions=a4)
    assert q.getType() == a1
    assert q.getVariable() == a2
    assert q.getWhere() == a3
    assert q.retrieveAllVersions() == a4

    # make a productStorage
    thepoolname = Test_Pool_Name
    thepoolurl = poolpath + '/' + thepoolname
    thepool, pstore = mkStorage(thepoolname, thepoolurl)

    # make another
    newpoolname = 'new_' + Test_Pool_Name
    newpoolurl = newpoolpath + '/' + newpoolname
    newpool, pstore2 = mkStorage(newpoolname, newpoolurl)

    # add some products to both storages
    n = 7
    rec1 = []
    for i in range(n):
        a0, a1, a2 = 'desc %d' % i, 'fatman %d' % (i*4), 5000+i
        if i < 3:
            x = TP(description=a0, creator=a1)
            x.meta['extra'] = Parameter(value=a2)
        elif i < 5:
            x = Context(description=a0, creator=a1)
            x.meta['extra'] = Parameter(value=a2)
        else:
            x = MapContext(description=a0, creator=a1)
            x.meta['extra'] = Parameter(value=a2)
            x.meta['time'] = Parameter(value=FineTime1(a2))
        if i < 4:
            r = pstore.save(x)
        else:
            r = pstore2.save(x)
        rec1.append(dict(p=x, r=r, a0=a0, a1=a1, a2=a2))

    # [T T T C] [C M M]
    #  0 1 2 3   4 5 6

    assert thepool.count == 4
    assert newpool.count == 3

    # query with a specific parameter in all products' metadata, which is the variable 'm' in the query expression, i.e. ``m = product.meta; ...``
    m = 2

    q = MetaQuery(TP, 'm["description"].value == "%s"' % rec1[m]['a0'])
    res = pstore.select(q)

    def chk(r, c):
        p = r.product
        assert type(p) == type(c['p'])
        assert p.description == c['a0']
        assert p.creator == c['a1']
        assert p.meta['extra'].value == c['a2']

    chk(res[0], rec1[m])

    # gracefully handle none-exisiting key
    # q = MetaQuery(TP, 'm["not_exists"].value == "%s"' % rec1[m]['a0'])
    # res = pstore.select(q)

    # query with a parent class and a specific parameter
    m = 3
    q = MetaQuery(BaseProduct, 'm["creator"].value == "%s"' % rec1[m]['a1'])
    res = pstore.select(q)
    assert len(res) == 1, str(res)
    chk(res[0], rec1[m])
    # query with a parent class and a specific parameter
    q = MetaQuery(BaseProduct, 'm["extra"].value < 5002')
    res = pstore.select(q)
    # [0,1]
    assert len(res) == 2, str(res)
    chk(res[0], rec1[0])
    chk(res[1], rec1[1])

    # simpler syntax for comparing value only but a bit slower.
    # the parameter with simpler syntax must be on the left hand side of a comparison operator.
    # '5000 < m["extra"]' does not work. But '5000 < m["extra"].value' works.
    q = MetaQuery(BaseProduct, 'm["extra"] > 5000 and m["extra"] <= 5002')
    res = pstore.select(q)
    # [1,2]
    assert len(res) == 2, str(res)
    chk(res[0], rec1[1])
    chk(res[1], rec1[2])

    qstr = 'm["extra"] > 5000 and m["extra"] <= 5002'
    pqv = pstore.select(qstr, variable='m', ptype=BaseProduct)
    pqs = pstore.select(qstr, variable='m', ptype='BaseProduct')
    assert res == pqv == pqs

    # two classes
    q = MetaQuery(BaseProduct, 'm["extra"] > 5000 and m["extra"] < 5004')
    res = pstore.select(q)
    # [1,2,3]
    assert len(res) == 3, str(res)
    chk(res[0], rec1[1])
    chk(res[1], rec1[2])
    chk(res[2], rec1[3])

    # this is not in this store
    q = MetaQuery(BaseProduct, 'm["extra"] == 5004')
    res = pstore.select(q)
    # []
    assert len(res) == 0, str(res)

    # it is in the other store
    q = MetaQuery(BaseProduct, 'm["extra"] == 5004')
    res = pstore2.select(q)
    # [4]
    assert len(res) == 1, str(res)
    chk(res[0], rec1[4])

    # all in  the other store
    q = MetaQuery(BaseProduct, '1')
    res = pstore2.select(q)
    # [4,5,6]
    assert len(res) == 3, str(res)
    chk(res[0], rec1[4])
    chk(res[1], rec1[5])
    chk(res[2], rec1[6])

    # register the new pool above to the  1st productStorage
    pstore.register(newpoolname)
    assert len(pstore.getPools()) == 2
    assert pstore.getPools()[1] == newpoolname

    # all Context, spans over two pools
    q = MetaQuery(Context, 'True')
    res = pstore.select(q)
    # [3,4,5,6]
    assert len(res) == 4, str(res)
    chk(res[0], rec1[3])
    chk(res[1], rec1[4])
    chk(res[2], rec1[5])
    chk(res[3], rec1[6])

    qstr = 'True'
    pqv = pstore.select(qstr, variable='m', ptype=Context)
    pqs = pstore.select(qstr, variable='m', ptype='Context')
    assert res == pqv == pqs

    # all 'time' < 5006. will cause KeyError because some Contex data do not have 'time'
    q = MetaQuery(Context, 'm["time"] < 5006')
    with pytest.raises(KeyError):
        res = pstore.select(q)

    q = MetaQuery(Context, 'm["this_and_the_last_errors_are_expected"] < 5006')
    with pytest.raises(KeyError):
        res = pstore.select(q)

    # all 'time' < 5006 mapcontext. all in newpool
    q = MetaQuery(MapContext, 'm["time"] < 5006')
    res = pstore.select(q)
    # [5]
    assert len(res) == 1, str(res)
    chk(res[0], rec1[5])

    # all 'extra' < 5002, all in 1st pool
    q = MetaQuery(BaseProduct, 'm["extra"] < 5002')
    res = pstore.select(q)
    # [0,1   ]
    assert len(res) == 2, str(res)
    chk(res[0], rec1[0])
    chk(res[1], rec1[1])

    # creator = 'fatman 12|16', two pools
    q = MetaQuery(BaseProduct, '"n 1" in m["creator"].value')
    res = pstore.select(q)
    # [3,4]
    assert len(res) == 2, str(res)
    chk(res[0], rec1[3])
    chk(res[1], rec1[4])

    qstr = '"n 1" in m["creator"].value'
    pq1 = thepool.select(qstr, ptype='BaseProduct')
    pq2 = newpool.qm(qstr, 'BaseProduct')
    assert res == pq1 + pq2

    if not newpoolpath.startswith('http'):
        # same as above but query is a function
        def t(m):
            import re
            return re.match('.*n.1.*', m['creator'].value)

        q = MetaQuery(BaseProduct, t)
        res = pstore.select(q)
        # [3,4]
        assert len(res) == 2, str(res)
        chk(res[0], rec1[3])
        chk(res[1], rec1[4])

    # same as above but query is on the product. this is slow.
    q = AbstractQuery(BaseProduct, 'p', '"n 1" in p.creator')
    res = pstore.select(q)
    # [3,4]
    assert len(res) == 2, str(res)
    chk(res[0], rec1[3])
    chk(res[1], rec1[4])

    qstr = '"n 1" in p.creator'
    pqv = pstore.select(qstr, variable='p', ptype=BaseProduct)
    pqs = pstore.select(qstr, variable='p', ptype='BaseProduct')
    assert res == pqv == pqs
    # pool
    pq1 = thepool.select(qstr, 'p', 'BaseProduct')
    pq2 = newpool.select(qstr, 'p', 'BaseProduct')
    assert res == pq1 + pq2

    q = '"n 1" in p.creator'
    resu = thepool.where(q)
    res = [ProductRef(u) for u in resu]
    # [3]
    assert len(res) == 1, str(res)
    chk(res[0], rec1[3])

    resu = thepool.where('p.meta["extra"] > 5000 and p.meta["extra"] < 5004')
    res = [ProductRef(u) for u in resu]
    # [1,2,3]
    assert len(res) == 3, str(res)
    chk(res[0], rec1[1])
    chk(res[1], rec1[2])
    chk(res[2], rec1[3])

    # report lru_cache info
    print('***cache*** %s ' % str(thepool.getCacheInfo()))


def test_query_local_mem():
    cleanup()
    thepoolpath = '/tmp/fditest'
    doquery('file://'+thepoolpath, 'file://'+thepoolpath)
    doquery('mem://'+thepoolpath, 'mem://'+thepoolpath)
    doquery('file://'+thepoolpath, 'mem://'+thepoolpath)
    doquery('mem://'+thepoolpath, 'file://'+thepoolpath)


def test_query_http(server):

    aburl, hdrs = server
    aburl = aburl.rstrip('/')
    cleanup()
    lpath = '/tmp'
    # TODO: http
    doquery(aburl, aburl)
    doquery('file://'+lpath, aburl)
    doquery('mem://'+lpath, aburl)


def test_RefContainer():
    # construction
    owner = Context(description='owner')
    v = RefContainer()
    v.setOwner(owner)
    assert v._owner == owner
    # add
    image = ProductRef(Product(description="hi"))
    assert len(image.parents) == 0
    v['i'] = image
    assert v.get('i') == image
    spectrum = ProductRef(Product(description="there"))
    v.put('s', spectrum)
    assert v['s'] == spectrum
    simple = ProductRef(Product(description="everyone"))
    v.set('m', simple)
    assert v.size() == 3
    # number of parents becomes 1
    assert len(image.parents) == 1
    # te parent is..
    assert spectrum.parents[0] == owner

    # del
    del v['s']
    assert 'i' in v
    assert 'm' in v
    assert 's' not in v
    assert len(v) == 2
    # no parent any more
    assert len(spectrum.parents) == 0

    checkjson(v)


@pytest.fixture(scope='function')
def a_storage():

    thepoolname = Test_Pool_Name
    thepoolpath = '/tmp/fditest'
    thepoolurl = 'file://' + thepoolpath + '/'+thepoolname
    # remove existing pools in memory
    PoolManager().removeAll()
    # create a product store
    pstore = ProductStorage(thepoolname, thepoolurl)
    assert len(pstore.getPools()) == 1
    assert pstore.getWritablePool() == thepoolname
    assert op.isdir(transpath(thepoolname, thepoolpath))
    # clean up possible garbage of previous runs
    pstore.wipePool()
    assert op.isdir(transpath(thepoolname, thepoolpath))
    assert sum([1 for x in glob.glob(
        op.join(transpath(thepoolname, thepoolpath), '*'))]) == 0
    return pstore, thepoolname, thepoolurl


def test_Context(a_storage):
    c1 = Context(description='1')
    c2 = Context(description='2')
    assert Context.isContext(c2.__class__)
    with pytest.raises(NotImplementedError):
        assert c1.isValid()

    # dirtiness
    # assert not c1.hasDirtyReferences('ok')
    #


def test_MapContext(a_storage):
    # doc
    image = Product(description="hi")
    spectrum = Product(description="there")
    simple = Product(description="everyone")

    context = MapContext()
    context.refs.put("x", ProductRef(image))
    context.refs.put("y", ProductRef(spectrum))
    context.refs.put("z", ProductRef(simple))
    assert context.refs.size() == 3
    assert context.refs.get('x').product.description == 'hi'
    assert context.refs.get('y').product.description == 'there'
    assert context.refs.get('z').product.description == 'everyone'

    product4 = Product(description="everybody")
    context.refs.put("y", ProductRef(product4))
    product5 = Product(description="here")
    context.refs.put("a", ProductRef(product5))

    assert context.refs.get('x').product.description == 'hi'
    assert context.refs.get('y').product.description == 'everybody'
    assert context.refs.get('z').product.description == 'everyone'
    assert context.refs.get('a').product.description == 'here'

    # access
    c1 = MapContext()
    # syntax 1. refs is a property to MapContext
    c1.refs.put("x", ProductRef(image))
    c2 = MapContext()
    # syntax 2  # put == set
    c2.refs.set("x", ProductRef(image))
    # assert c1 == c2, deepcmp(c1, c2)
    c3 = MapContext()
    # syntax 3 # refs is a composite so set/get = []
    c3.refs["x"] = ProductRef(image)
    # assert c3 == c2
    assert c3.refs['x'].product.description == 'hi'
    c4 = MapContext()
    # syntax 4. refs is a member in a composite (Context) so set/get = []
    c4['refs']["x"] = ProductRef(image)
    # assert c3 == c4
    assert c4['refs']['x'].product.description == 'hi'

    # stored prod
    pstore, thepoolname, purl = a_storage
    # create a prooduct
    x = Product(description='save me in store')
    # save the product and get a reference
    prodref = pstore.save(x)
    # has the ProductStorage
    assert prodref.getPoolname() == thepoolname
    # has the pool
    assert prodref._poolname == thepoolname
    # returns the product
    assert prodref.product == x
    # create an empty mapcontext
    mc = MapContext()
    # put the ref in the context.
    # The manual has this syntax mc.refs.put('xprod', prodref)
    # but I like this for doing the same thing:
    mc['refs']['xprod'] = prodref
    # get the urn
    urn = prodref.urn
    assert issubclass(urn.__class__, str)
    # re-create a product only using the urn
    newp = ProductRef(urn).product
    # the new and the old one are equal
    assert newp == x
    # parent is set
    assert prodref.parents[0] == mc
    # re-create a product only using the urn 2
    newref = pstore.load(urn)
    newp2 = newref.product
    # the new and the old one are equal
    assert newp2 == x

    des = checkjson(mc, dbg=0)
    # print(type(des['refs']))
    # print('&&&&&& ' + des.refs.serialized(indent=4) + ' %%%%%%')
    # print(yaml.dump(des))

    newx = des['refs']['xprod'].product
    assert newx == x

    # remove refs
    del mc.refs['xprod']
    assert mc.refs.size() == 0
    assert len(prodref.parents) == 0
    # another way to remove
    des.refs.pop('xprod')
    assert des.refs.size() == 0
    assert len(prodref.parents) == 0
    # clear all
    prodref2 = pstore.save(Product())
    mc.refs['a'] = prodref
    mc.refs['b'] = prodref2
    assert mc.refs.size() == 2
    mc.refs.clear()
    assert mc.refs.size() == 0

    # URN of an object in memory
    urn = ProductRef(x).urn
    newp = PoolManager.getPool(DEFAULT_MEM_POOL).loadProduct(urn)
    # the new and the old one are equal
    assert newp == x

    c1 = MapContext(description='1')
    c2 = MapContext(description='2')
    # getAllRefs
    pstore, thepoolname, purl = a_storage
    p1 = Product('p1')
    p2 = Product('p2')
    # _urn for product is set.
    assert not hasattr(p1, '_urn')
    refp1 = pstore.save(p1)
    assert p1._urn == refp1.urnobj
    refp2 = pstore.save(p2)
    c1.refs['p1'] = refp1
    refc1 = pstore.save(c1)
    c2.refs['p2'] = refp2
    c2.refs['c1'] = refc1
    assert c2.getAllRefs() == [refp2, refc1]
    assert c2.getAllRefs(includeContexts=False) == [refp2]
    assert c2.getAllRefs(recursive=True) == [refp2, refc1, refp1]

    # realistic scenario


def test_realistic_http(server, demo_product):

    aburl, hdrs = server
    aburl = aburl.rstrip('/')
    poolname = 'demo'
    poolurl = aburl + '/' + poolname
    cleanup(poolurl, poolname)
    # remove existing pools in memory
    # clean up possible garbage of previous runs. use class method to avoid reading pool hk info during ProdStorage initialization.
    thepool, pstore = mkStorage(poolname, poolurl)
    pass


def test_realistic_csdb(csdb, urlcsdb, csdb_client, demo_product):
    urlupload, urldelete, urllist, client = csdb_client
    test_pool, poolurl, pstore = csdb  # csdb:///csdb_test_pool
    poolname = test_pool

    cleanup(poolurl, poolname)
    # remove existing pools in memory
    # clean up possible garbage of previous runs. use class method to avoid reading pool hk info during ProdStorage initialization.
    thepool, pstore = mkStorage(poolname, poolurl, pstore)

    p1 = Product(description='p1')
    p2 = Product(description='p2')
    map1 = MapContext(description='product with refs 1')
    # A ProductRef created from a lone product will use a mempool
    pref1 = ProductRef(p1)
    # use a productStorage with a pool on disk
    pref2 = pstore.save(p2)
    # how many prodrefs do we have? (do not use len() due to _STID, version)
    assert map1['refs'].size() == 0
    assert len(pref1.parents) == 0
    assert len(pref2.parents) == 0
    # add a ref to the contex. every ref has a name in mapcontext
    map1['refs']['spam'] = pref1
    assert map1['refs'].size() == 1
    assert len(pref1.parents) == 1
    assert pref1.parents[0] == map1
    # add the second one
    map1['refs']['egg'] = pref2
    # how many prodrefs do we have? (do not use len() due to _STID, version)
    assert map1['refs'].size() == 2
    assert len(pref2.parents) == 1
    assert pref2.parents[0] == map1
    assert pref1.parents[0] == map1
    # remove a ref
    del map1['refs']['spam']
    # how many prodrefs do we have? (do not use len() due to _STID, version)
    assert map1.refs.size() == 1
    assert len(pref1.parents) == 0
    # add ref2 to another map
    map2 = MapContext(description='product with refs 2')
    map2.refs['also2'] = pref2
    assert map2['refs'].size() == 1
    # two parents
    assert len(pref2.parents) == 2
    assert pref2.parents[1] == map2
    map1ref = pstore.save(map1, tag=__name__+'::realistic.map1')
    map1ref = pstore.save(map2, tag=__name__+'::realistic.map2')

    # demo prod
    dp, related = demo_product
    relatedref = pstore.save(related, tag='referenced by DemoProduct')
    dp['refs']['a related product'] = relatedref
    demoprodref = pstore.save(dp, tag='DemoProduct')


def f(n):
    return list(itertools.repeat(random.random(), n))


def rands(n):
    return [random.random() for i in range(n)]


def h(n):
    return [random.random()] * n


def speed():
    m = 10000
    print(timeit.timeit('[func(%d) for func in (rands,)]' % m,
                        globals=globals(), number=1))
    a = ArrayDataset(rands(m))
    p = Product(description="product example",
                instrument="Favourite", modelName="Flight")
    p['array'] = a
    PoolManager().removeAll()
    # create a product store
    pool = 'file:///tmp/perf_' + Test_Pool_Name
    pstore = ProductStorage(pool)
    # clean up possible garbage of previous runs
    pstore.wipePool()
    # in memory
    print(timeit.timeit('ref1 = ProductRef(p)',
                        globals=globals().update(locals()), number=1))
    pref2 = pstore.save(p)  # on disk


def running(t):
    print('running ' + str(t))
    t()


if __name__ == '__main__' and __package__ is None:
    speed()
    exit()
    running(test_ProductRef)
    running(test_ProductRef)
    running(test_MapContext)
    running(test_Urn)
    # running(test_MapRefsDataset)
    running(test_PoolManager)
    running(test_ProductStorage)
    running(test_Context)
